############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries
import logging

# external packages
import PyQt5

# local imports


class MeasureDataRaw(PyQt5.QtCore.QObject):
    """
    the class MeasureData inherits all information and handling of data management and
    storage

        >>> measure = MeasureDataRaw(
        >>>             app=None,
        >>>             parent=None,
        >>>             data=None,
        >>>                 )
    """

    __all__ = ['MeasureDataRaw',
               ]

    log = logging.getLogger(__name__)

    # update rate to 1 seconds for setting indi server
    CYCLE_UPDATE_TASK = 1000
    # maximum size of measurement task
    MAXSIZE = 24 * 60 * 60

    def __init__(self, app=None, parent=None, data=None):
        super().__init__()

        self.app = app
        self.parent = parent
        self.data = data
        self.deviceName = 'RAW'
        self.defaultConfig = {
            'raw': {
                'deviceName': 'display only',
            }
        }

    # time for measurement
        self.timerTask = PyQt5.QtCore.QTimer()
        self.timerTask.setSingleShot(False)
        self.timerTask.timeout.connect(self.measureTask)

    def startCommunication(self, loadConfig=False):
        """
        startCommunication starts cycling of the polling.

        :param loadConfig:
        :return: True for test purpose
        """

        self.parent.setEmptyData()
        self.timerTask.start(self.CYCLE_UPDATE_TASK)

        return True

    def stopCommunication(self):
        """
        stopCommunication stops the devices

        :return: true for test purpose
        """

        self.timerTask.stop()
        return True

    def measureTask(self):
        """
        measureTask runs all necessary pre processing and collecting task to assemble a
        large dict of lists, where all measurement data is stored. the intention later on
        would be to store and export this data.
        the time object is related to the time held in mount computer and is in utc timezone.

        data sources are:
            environment
            mount pointing position

        :return: success
        """

        suc = self.parent.measureTask()

        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import logging

# external packages
import skyfield.api as api
import skyfield.starlib as starlib

# local imports
from .connection import Connection
from .convert import stringToAngle
from .convert import valueToFloat
from .convert import valueToInt
from .convert import valueToAngle
from .convert import sexagesimalizeToInt


class ObsSite(object):
    """
    The class Site inherits all information and handling of site data
    attributes of the connected mount and provides the abstracted interface
    to a 10 micron mount. as the mount's time base is julian date, we use this
    value as time base as well. for that reason we should remind how the mount
    calculates the julian date. it is derived from utc. to basically on the
    timeJD for skyfield we calculate julian date from ut1 based on julian date
    from mount based on utc and the value delta utc, ut1 also given from the
    mount.

        >>> site = ObsSite(
        >>>             host=host,
        >>>             pathToData=pathToData,
        >>>             verbose=verbose,
        >>>             location=(0, 0, 0),
        >>>             )

    The Site class needs as parameter a ts object from skyfield.api to
    be able to make all the necessary calculations about time from and to mount
    """

    __all__ = ['ObsSite',
               ]

    log = logging.getLogger(__name__)

    STAT = {
        '0': 'Tracking',
        '1': 'Stopped after STOP',
        '2': 'Slewing to park position',
        '3': 'Unparking',
        '4': 'Slewing to home position',
        '5': 'Parked',
        '6': 'Slewing or going to stop',
        '7': 'Tracking Off no move',
        '8': 'Motor low temperature',
        '9': 'Tracking outside limits',
        '10': 'Following Satellite',
        '11': 'User OK Needed',
        '98': 'Unknown Status',
        '99': 'Error',
    }

    def __init__(self,
                 host=None,
                 pathToData=None,
                 verbose=False,
                 ):

        self.host = host
        self.pathToData = pathToData
        self.verbose = verbose
        self.loader = None
        self._location = None
        self.ts = None
        self._timeJD = None
        self._utc_ut1 = None
        self._timeSidereal = None
        self._raJNow = None
        self._raJNowTarget = None
        self._decJNow = None
        self._decJNowTarget = None
        self._haJNow = None
        self._haJNowTarget = None
        self._sidereal = None
        self._angularPosRA = None
        self._angularPosDEC = None
        self._angularPosRATarget = None
        self._angularPosDECTarget = None
        self._pierside = None
        self._piersideTarget = None
        self._Alt = None
        self._AltTarget = None
        self._Az = None
        self._AzTarget = None
        self._status = None
        self._statusSlew = None
        self.setLoaderAndTimescale()

    def setLoaderAndTimescale(self):
        """
        :return:
        """
        if self.pathToData:
            self.loader = api.Loader(self.pathToData, verbose=self.verbose)
        else:
            self.loader = api.load

        self.ts = self.loader.timescale(builtin=True)
        self.log.debug('Timescale is using built-in')
        return True

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        if isinstance(value, api.Topos):
            self._location = value
            return

        if not isinstance(value, (list, tuple)):
            self._location = None
            self.log.info(f'Malformed value: {value}')
            return

        if len(value) != 3:
            self._location = None
            self.log.info(f'Malformed value: {value}')
            return

        lat, lon, elev = value
        if not isinstance(lat, api.Angle):
            lat = stringToAngle(lat, preference='degrees')

        if not isinstance(lon, api.Angle):
            lon = stringToAngle(lon, preference='degrees')

        elev = valueToFloat(elev)
        if lat is None or lon is None or elev is None:
            self._location = None
            self.log.info(f'Malformed value: {value}')
            return

        self._location = api.Topos(longitude=lon,
                                   latitude=lat,
                                   elevation_m=elev)

    @property
    def timeJD(self):
        if self._timeJD is None:
            return self.ts.now()
        else:
            return self._timeJD

    @timeJD.setter
    def timeJD(self, value):
        value = valueToFloat(value)
        if value and self._utc_ut1 is not None:
            self._timeJD = self.ts.ut1_jd(value - self._utc_ut1)
        else:
            self._timeJD = None

    @property
    def utc_ut1(self):
        return self._utc_ut1

    @utc_ut1.setter
    def utc_ut1(self, value):
        value = valueToFloat(value)
        if value is not None:
            self._utc_ut1 = value / 86400
        else:
            self._utc_ut1 = None

    @property
    def timeSidereal(self):
        return self._timeSidereal

    @timeSidereal.setter
    def timeSidereal(self, value):
        if isinstance(value, str):
            self._timeSidereal = stringToAngle(value, preference='hours')
        elif isinstance(value, float):
            self._timeSidereal = valueToAngle(value, preference='hours')
        elif isinstance(value, api.Angle):
            self._timeSidereal = value
        else:
            self._timeSidereal = None

    @property
    def raJNow(self):
        return self._raJNow

    @raJNow.setter
    def raJNow(self, value):
        if isinstance(value, api.Angle):
            self._raJNow = value
            return
        self._raJNow = valueToAngle(value, preference='hours')

    @property
    def raJNowTarget(self):
        return self._raJNowTarget

    @raJNowTarget.setter
    def raJNowTarget(self, value):
        if isinstance(value, api.Angle):
            self._raJNowTarget = value
            return
        self._raJNowTarget = stringToAngle(value, preference='hours')

    @property
    def haJNow(self):
        if self._timeSidereal is None or self._raJNow is None:
            return None
        else:
            # ha is always positive between 0 and 24 hours
            ha = (self._timeSidereal.hours - self._raJNow.hours + 24) % 24
            return api.Angle(hours=ha)

    @property
    def haJNowTarget(self):
        if self._timeSidereal is None or self._raJNowTarget is None:
            return None
        else:
            # ha is always positive between 0 and 24 hours
            ha = (self._timeSidereal.hours - self._raJNowTarget.hours + 24) % 24
            return api.Angle(hours=ha)

    @property
    def decJNow(self):
        return self._decJNow

    @decJNow.setter
    def decJNow(self, value):
        if isinstance(value, api.Angle):
            self._decJNow = value
            return
        self._decJNow = valueToAngle(value, preference='degrees')

    @property
    def decJNowTarget(self):
        return self._decJNowTarget

    @decJNowTarget.setter
    def decJNowTarget(self, value):
        if isinstance(value, api.Angle):
            self._decJNowTarget = value
            return
        self._decJNowTarget = stringToAngle(value, preference='degrees')

    @property
    def angularPosRA(self):
        return self._angularPosRA

    @angularPosRA.setter
    def angularPosRA(self, value):
        if isinstance(value, api.Angle):
            self._angularPosRA = value
            return
        self._angularPosRA = valueToAngle(value, preference='degrees')

    @property
    def angularPosDEC(self):
        return self._angularPosDEC

    @angularPosDEC.setter
    def angularPosDEC(self, value):
        if isinstance(value, api.Angle):
            self._angularPosDEC = value
            return
        self._angularPosDEC = valueToAngle(value, preference='degrees')

    @property
    def angularPosRATarget(self):
        return self._angularPosRATarget

    @angularPosRATarget.setter
    def angularPosRATarget(self, value):
        if isinstance(value, api.Angle):
            self._angularPosRATarget = value
            return
        self._angularPosRATarget = valueToAngle(value, preference='degrees')

    @property
    def angularPosDECTarget(self):
        return self._angularPosDECTarget

    @angularPosDECTarget.setter
    def angularPosDECTarget(self, value):
        if isinstance(value, api.Angle):
            self._angularPosDECTarget = value
            return
        self._angularPosDECTarget = valueToAngle(value, preference='degrees')

    @property
    def pierside(self):
        return self._pierside

    @pierside.setter
    def pierside(self, value):
        if value in ['E', 'W', 'e', 'w']:
            value = value.capitalize()
            self._pierside = value
        else:
            self._pierside = None
            self.log.info(f'Malformed value: {value}')

    @property
    def piersideTarget(self):
        return self._piersideTarget

    @piersideTarget.setter
    def piersideTarget(self, value):
        if value == '2':
            self._piersideTarget = 'W'
        elif value == '3':
            self._piersideTarget = 'E'
        else:
            self._piersideTarget = None
            self.log.info(f'Malformed value: {value}')

    @property
    def Alt(self):
        return self._Alt

    @Alt.setter
    def Alt(self, value):
        if isinstance(value, api.Angle):
            self._Alt = value
            return
        self._Alt = valueToAngle(value, preference='degrees')

    @property
    def AltTarget(self):
        return self._AltTarget

    @AltTarget.setter
    def AltTarget(self, value):
        if isinstance(value, api.Angle):
            self._AltTarget = value
            return
        self._AltTarget = stringToAngle(value, preference='degrees')

    @property
    def Az(self):
        return self._Az

    @Az.setter
    def Az(self, value):
        if isinstance(value, api.Angle):
            self._Az = value
            return
        self._Az = valueToAngle(value, preference='degrees')

    @property
    def AzTarget(self):
        return self._AzTarget

    @AzTarget.setter
    def AzTarget(self, value):
        if isinstance(value, api.Angle):
            self._AzTarget = value
            return
        self._AzTarget = stringToAngle(value, preference='degrees')

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = valueToInt(value)
        if self._status not in [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 98, 99]:
            self._status = None

    def statusText(self):
        if self._status is None:
            return None
        reference = f'{self._status:d}'
        if reference in self.STAT:
            return self.STAT[reference]

    @property
    def statusSlew(self):
        return self._statusSlew

    @statusSlew.setter
    def statusSlew(self, value):
        self._statusSlew = bool(value)

    def parseLocation(self, response, numberOfChunks):
        """
        Parsing the polling slow command.

        :param response:        data load from mount
        :param numberOfChunks:  amount of parts
        :return: success:       True if ok, False if not
        """

        if len(response) != numberOfChunks:
            self.log.warning('Wrong number of chunks')
            return False
        elev = response[0]
        # due to compatibility to LX200 protocol east is negative, so we change that
        # in class we would like to keep the correct sign for east is positive
        lon = None
        if '-' in response[1]:
            lon = response[1].replace('-', '+')
        if '+' in response[1]:
            lon = response[1].replace('+', '-')
        lat = response[2]

        self.location = [lat, lon, elev]
        return True

    def getLocation(self):
        """
        Sending the polling command. As the mount need polling the data, I send
        a set of commands to get the data back to be able to process and store it.

        :return: success:   True if ok, False if not
        """

        conn = Connection(self.host)
        commandString = ':U2#:Gev#:Gg#:Gt#'
        suc, response, numberOfChunks = conn.communicate(commandString)
        if not suc:
            return False
        suc = self.parseLocation(response, numberOfChunks)
        return suc

    def parsePointing(self, response, numberOfChunks):
        """
        Parsing the polling fast command.

        :param response:        data load from mount
        :param numberOfChunks:  amount of parts
        :return: success:       True if ok, False if not
        """

        if len(response) != numberOfChunks:
            self.log.warning('Wrong number of chunks')
            return False
        self.timeSidereal = response[0]
        # remove the leap seconds flag if present
        self.utc_ut1 = response[1].replace('L', '')
        self.angularPosRA = response[2]
        self.angularPosDEC = response[3]
        responseSplit = response[4].split(',')
        self.raJNow = responseSplit[0]
        self.decJNow = responseSplit[1]
        self.pierside = responseSplit[2]
        self.Az = responseSplit[3]
        self.Alt = responseSplit[4]
        self.timeJD = responseSplit[5]
        self.status = responseSplit[6]
        self.statusSlew = (responseSplit[7] == '1')
        return True

    def pollPointing(self):
        """
        Sending the polling fast command. As the mount need polling the data, I send
        a set of commands to get the data back to be able to process and store it.

        :return: success:   True if ok, False if not
        """

        conn = Connection(self.host)
        commandString = ':U2#:GS#:GDUT#:GaXa#:GaXb#:Ginfo#'
        suc, response, numberOfChunks = conn.communicate(commandString)
        if not suc:
            return False
        suc = self.parsePointing(response, numberOfChunks)
        return suc

    def startSlewing(self, slewType='normal', forceUnpark=False):
        """
        startSlewing issues the final slew command to the mount after the target
        coordinates were set. before issuing the slewing command it automatically unpark
        the mount as well.

        the slew commands are:
            :MS#  :MA#  :MSap#  :MSao# :MaX#, :PiP#

        and return:
            0 no error
                if the target is below the lower limit: the string
            “1Object Below Horizon #”
                if the target is above the high limit: the string
            “2Object Below Higher #”
                if the slew cannot be performed due to another cause: the string
            “3Cannot Perform Slew #”
                if the mount is parked: the string
            “4Mount Parked #”
                if the mount is restricted to one side of the meridian and the object
                is on the other side: the string
            “5Object on the other side #”

        the types of slew is:
        - 'normal'      slew to coordinates and tracking on
        - 'notrack':    slew to coordinates and tracking off
        - 'stop':       slew to coordinates and park
        - 'park':       slew to coordinates and park
        - 'polar':      slew to coordinates and miss for polar alignment
        - 'ortho':      slew to coordinates and miss for orthogonal alignment
        - 'keep':       choose between normal and notrack to keep the tracking mode


        :param slewType:
        :return:
        """
        slewTypes = {
            'normal': ':MS#',
            'notrack': ':MA#',
            'stop': ':MaX#',
            'park': ':PaX#',
            'polar': ':MSap#',
            'ortho': ':MSao#',
            'keep': '',
        }

        if slewType not in slewTypes:
            return False

        if slewType == 'keep':
            if self.status == 0:
                slewTypes['keep'] = ':MS#'

            else:
                slewTypes['keep'] = ':MA#'

        conn = Connection(self.host)

        commandString = ':PO#' + slewTypes[slewType]
        suc, response, numberOfChunks = conn.communicate(commandString)
        if not suc:
            return False

        if not response[0].startswith('0'):
            self.log.debug(f'Slew could not be done: [{response}]')
            return False

        return True

    def setTargetAltAz(self,
                       alt=None, az=None,
                       alt_degrees=None, az_degrees=None):
        """
        Slew AltAz unpark the mount sets the targets for alt and az and then
        issue the slew command.

        the unpark command is:
            :PO#
        and returns nothing

        setting alt target is the following:
            :SzDDD*MM# or :SzDDD*MM:SS# or :SzDDD*MM:SS.S#, we use the last one
            :SzDDD*MM:SS.S#

        setting az target is the following:
            :SasDD*MM# or :SasDD*MM:SS# or :SasDD*MM:SS.S#, we use the last one
            :SasDD*MM:SS.S#

        the slew command moves the mount and keeps tracking at the end of the move.
        in the command protocol it is written, that the targets should be ra / dec,
        but it works for targets defined with alt / az commands

        :param alt:     altitude in type Angle
        :param az:      azimuth in type Angle
        :param alt_degrees:     altitude in degrees float
        :param az_degrees:      azimuth in degrees float
        :return:        success
        """
        hasAngles = isinstance(alt, api.Angle) and isinstance(az, api.Angle)
        altHasFloat = isinstance(alt_degrees, (float, int))
        azHasFloat = isinstance(az_degrees, (float, int))

        if hasAngles:
            pass

        elif altHasFloat and azHasFloat:
            alt = api.Angle(degrees=alt_degrees)
            az = api.Angle(degrees=az_degrees)

        else:
            return False

        if alt.preference != 'degrees':
            return False

        if az.preference != 'degrees':
            return False

        conn = Connection(self.host)

        sgn, h, m, s, frac = sexagesimalizeToInt(alt.degrees, 1)
        sign = '+' if sgn >= 0 else '-'
        setAlt = f':Sa{sign}{h:02d}*{m:02d}:{s:02d}.{frac:1d}#'

        sgn, h, m, s, frac = sexagesimalizeToInt(az.degrees, 1)
        sign = '+' if sgn >= 0 else '-'
        setAz = f':Sz{sign}{h:02d}*{m:02d}:{s:02d}.{frac:1d}#'

        getTargetStatus = ':U2#:GTsid#:Ga#:Gz#:Gr#:Gd#'
        commandString = setAlt + setAz + getTargetStatus
        suc, response, numberOfChunks = conn.communicate(commandString)
        if not suc:
            return False

        result = response[0][0:2]
        if result.count('0') > 0:
            self.log.debug(f'Coordinates could not be set: [{response}]')
            return False

        if len(response) != 4:
            self.log.debug(f'Missing return values: [{response}]')
            return False

        self.piersideTarget = response[0][2]
        self.AltTarget = response[0][3:]
        self.AzTarget = response[1]
        self.raJNowTarget = response[2]
        self.decJNowTarget = response[3]
        return suc

    def setTargetRaDec(self,
                       ra=None, dec=None,
                       ra_hours=None, dec_degrees=None,
                       target=None):
        """
        Slew RaDec unpark the mount sets the targets for ra and dec and then
        issue the slew command.

        the unpark command is:
            :PO#
        and returns nothing

        setting ra target is the following:
            :SrHH:MM.T# or :SrHH:MM:SS# or :SrHH:MM:SS.S# or :SrHH:MM:SS.SS#
                , we use the last one
            :SrHH:MM:SS.SS#

        setting dec target is the following:
            :SdsDD*MM# or :SdsDD*MM:SS# or :Sd sDD*MM:SS.S#, we use the last one
            :SdsDD*MM:SS.S#

        the slew command moves the mount and keeps tracking at the end of the move.
        in the command protocol it is written, that the targets should be ra / dec,
        but it works for targets defined with alt / az commands

        :param ra:     right ascension in type Angle
        :param dec:    declination in type Angle preference 'hours'
        :param ra_hours: right ascension in float hours
        :param dec_degrees: declination in float degrees
        :param target:  star in type skyfield.Star
        :return:       success
        """
        hasTarget = isinstance(target, starlib.Star)
        hasAngles = isinstance(ra, api.Angle) and isinstance(dec, api.Angle)
        raHasFloat = isinstance(ra_hours, (float, int))
        decHasFloat = isinstance(dec_degrees, (float, int))
        if hasTarget:
            ra = target.ra
            dec = target.dec

        elif hasAngles:
            pass

        elif raHasFloat and decHasFloat:
            ra = api.Angle(hours=ra_hours, preference='hours')
            dec = api.Angle(degrees=dec_degrees)

        else:
            return False

        if ra.preference != 'hours' or dec.preference != 'degrees':
            return False

        conn = Connection(self.host)

        sgn, h, m, s, frac = sexagesimalizeToInt(ra.hours, 2)
        setRa = f':Sr{h:02d}:{m:02d}:{s:02d}.{frac:02d}#'

        sgn, h, m, s, frac = sexagesimalizeToInt(dec.degrees, 1)
        sign = '+' if sgn >= 0 else '-'
        setDec = f':Sd{sign}{h:02d}*{m:02d}:{s:02d}.{frac:1d}#'

        getTargetStatus = ':U2#:GTsid#:Ga#:Gz#:Gr#:Gd#'
        commandString = setRa + setDec + getTargetStatus
        suc, response, numberOfChunks = conn.communicate(commandString)
        if not suc:
            return False

        result = response[0][0:2]
        if result.count('0') > 0:
            self.log.debug(f'Coordinates could not be set: [{response}]')
            return False

        if len(response) != 4:
            self.log.debug(f'Missing return values: [{response}]')
            return False

        self.piersideTarget = response[0][2]
        self.AltTarget = response[0][3:]
        self.AzTarget = response[1]
        self.raJNowTarget = response[2]
        self.decJNowTarget = response[3]
        return suc

    def setTargetAngular(self,
                         ra=None, dec=None,
                         ra_degrees=None, dec_degrees=None,
                         target=None):
        """
        Slew RaDec unpark the mount sets the targets for ra and dec and then
        issue the slew command.

        the unpark command is:
            :PO#
        and returns nothing

        setting angular RA target is the following:
            :SaXasXXX.XXXX#

        setting angular DEC target is the following:
            ::SaXbsXXX.XXXX#

        :param ra:     right ascension in type Angle
        :param dec:    declination in type Angle preference 'hours'
        :param ra_degrees: right ascension in float degrees
        :param dec_degrees: declination in float degrees
        :param target:  star in type skyfield.Star
        :return:       success
        """
        hasTarget = isinstance(target, starlib.Star)
        hasAngles = isinstance(ra, api.Angle) and isinstance(dec, api.Angle)
        decHasFloat = isinstance(ra_degrees, (float, int))
        raHasFloat = isinstance(dec_degrees, (float, int))
        if hasTarget:
            ra = target.ra
            dec = target.dec

        elif hasAngles:
            pass

        elif raHasFloat and decHasFloat:
            ra = api.Angle(hours=ra_degrees)
            dec = api.Angle(degrees=dec_degrees)

        else:
            return False

        if dec.preference != 'degrees' or ra.preference != 'degrees':
            return False

        conn = Connection(self.host)
        raCommand = f':SaXa{ra.degrees:+03.4f}#'
        decCommand = f':SaXb{dec.degrees:+03.4f}#'
        getTargetStatus = ':U2#:GTsid#:Ga#:Gz#:Gr#:Gd#'
        commandString = raCommand + decCommand + getTargetStatus
        suc, response, numberOfChunks = conn.communicate(commandString)
        if not suc:
            return False

        result = response[0][0:2]
        if result.count('0') > 0:
            self.log.debug(f'Coordinates could not be set: [{response}]')
            return False

        if len(response) != 4:
            self.log.debug(f'Missing return values: [{response}]')
            return False

        # todo: actually SaXa and SaXb commands seem no to set other targets
        # self.piersideTarget = response[0][2]
        # self.AltTarget = response[0][3:]
        # self.AzTarget = response[1]
        # self.raJNowTarget = response[2]
        # self.decJNowTarget = response[3]
        return suc

    def shutdown(self):
        """
        shutdown send the shutdown command to the mount. if succeeded it takes about 20
        seconds before you could switch off the power supply. please check red LED at mount

        :return:    success
        """

        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':shutdown#')
        if not suc:
            return False
        if response[0] == '0':
            return False
        return True

    def setLocation(self, obs):
        """
        set SiteCoordinates sets the value in the mount to the given parameters.
        the longitude will be set for east negative, because that's the definition
        from the LX200 protocol, which the 10micron mount supports. internally we use
        the standard east positive.

        the site parameters could be set be the following commands:

        longitude (we will use the last one with the highest precision):

            :SgsDDD*MM# or :SgsDDD*MM:SS# or :SgsDDD*MM:SS.S#

            Set current site’s longitude to sDDD*MM (sign, degrees, arcminutes),
            sDDD*MM:SS (sign, degrees, arcminutes, arcseconds) or
            sDDD*MM:SS.S (sign, degrees, arcminutes, arcseconds and tenths of arcsecond).
            Note: East Longitudes are expressed as negative.
            Returns:
            0   invalid
            1   valid

        latitude (we will use the last one with the highest precision):

            :StsDD*MM# or :StsDD*MM:SS# or :StsDD*MM:SS.S#

            Sets the current site latitude to sDD*MM (sign, degrees, arcminutes),
            sDD*MM:SS (sign, degrees, arcminutes, arcseconds), or
            sDD*MM:SS.S (sign, degrees, arcminutes, arcseconds and tenths of arcsecond)

            Returns:
            0   invalid
            1   valid

        elevation:

            :SevsXXXX.X#

            Set current site’s elevation to sXXXX.X (sign, metres) in the
            range -1000.0 to 9999.9.
            Returns:
            0   invalid
            1   valid

        :param      obs:        skyfield.api.Topos of site
        :return:    success
        """
        if not isinstance(obs, api.Topos):
            return False

        conn = Connection(self.host)

        sgn, h, m, s, frac = sexagesimalizeToInt(obs.longitude.degrees, 1)
        sign = '+' if sgn < 0 else '-'
        setLon = f':Sg{sign}{h:03d}*{m:02d}:{s:02d}.{frac:1d}#'

        sgn, h, m, s, frac = sexagesimalizeToInt(obs.latitude.degrees, 1)
        sign = '+' if sgn >= 0 else '-'
        setLat = f':St{sign}{h:02d}*{m:02d}:{s:02d}.{frac:1d}#'

        sign = '+' if obs.elevation.m > 0 else '-'
        setElev = f':Sev{sign}{obs.elevation.m:06.1f}#'

        commandString = setLon + setLat + setElev
        suc, response, numberOfChunks = conn.communicate(commandString)
        if not suc:
            return False

        if '0' in response[0]:
            return False

        return True

    def setLatitude(self, lat=None, lat_degrees=None):
        """
        setLatitude sets the value in the mount to the given parameters.
        the site parameters could be set be the following commands:
        latitude (we will use the last one with the highest precision):

            :StsDD*MM# or :StsDD*MM:SS# or :StsDD*MM:SS.S#

            Sets the current site latitude to sDD*MM (sign, degrees, arcminutes),
            sDD*MM:SS (sign, degrees, arcminutes, arcseconds), or
            sDD*MM:SS.S (sign, degrees, arcminutes, arcseconds and tenths of arcsecond)

            Returns:
            0   invalid
            1   valid

        :param      lat:  coordinates as Angle
        :param      lat_degrees:  coordinates as float
        :return:    success
        """
        hasAngle = isinstance(lat, api.Angle)
        hasFloat = isinstance(lat_degrees, (float, int))
        hasStr = isinstance(lat, str)
        if hasAngle:
            pass

        elif hasFloat:
            lat = valueToAngle(lat_degrees, preference='degrees')

        elif hasStr:
            lat = lat.replace('deg', '')
            lat = stringToAngle(lat, preference='degrees')

        else:
            return False

        conn = Connection(self.host)

        sgn, h, m, s, frac = sexagesimalizeToInt(lat.degrees, 1)
        sign = '+' if sgn >= 0 else '-'
        setLat = f':St{sign}{h:02d}*{m:02d}:{s:02d}.{frac:1d}#'

        commandString = setLat
        suc, response, numberOfChunks = conn.communicate(commandString)
        if not suc:
            return False

        if '0' in response[0]:
            return False

        return True

    def setLongitude(self, lon=None, lon_degrees=None):
        """
        setLongitude sets the value in the mount to the given parameters.
        the longitude will be set for east negative, because that's the definition
        from the LX200 protocol, which the 10micron mount supports. internally we use
        the standard east positive.

        the site parameters could be set be the following commands:
        longitude (we will use the last one with the highest precision):

            :SgsDDD*MM# or :SgsDDD*MM:SS# or :SgsDDD*MM:SS.S#

            Set current site’s longitude to sDDD*MM (sign, degrees, arcminutes),
            sDDD*MM:SS (sign, degrees, arcminutes, arcseconds) or
            sDDD*MM:SS.S (sign, degrees, arcminutes, arcseconds and tenths of arcsecond).
            Note: East Longitudes are expressed as negative.
            Returns:
            0   invalid
            1   valid

        :param      lon:  coordinates as Angle
        :param      lon_degrees:  coordinates as float
        :return:    success
        """
        hasAngle = isinstance(lon, api.Angle)
        hasFloat = isinstance(lon_degrees, (float, int))
        hasStr = isinstance(lon, str)
        if hasAngle:
            pass

        elif hasFloat:
            lon = valueToAngle(lon_degrees, preference='degrees')

        elif hasStr:
            lon = lon.replace('deg', '')
            lon = stringToAngle(lon, preference='degrees')

        else:
            return False

        conn = Connection(self.host)

        sgn, h, m, s, frac = sexagesimalizeToInt(lon.degrees, 1)
        sign = '+' if sgn < 0 else '-'
        setLon = f':Sg{sign}{h:03d}*{m:02d}:{s:02d}.{frac:1d}#'

        commandString = setLon
        suc, response, numberOfChunks = conn.communicate(commandString)
        if not suc:
            return False

        if '0' in response[0]:
            return False

        return True

    def setElevation(self, elev):
        """
        setElevation sets the value in the mount to the given parameters.
        the site parameters could be set be the following commands:
        elevation:

            :SevsXXXX.X#

            Set current site’s elevation to sXXXX.X (sign, metres) in the
            range -1000.0 to 9999.9.
            Returns:
            0   invalid
            1   valid

        :param      elev:        string with elevation in meters
        :return:    success
        """
        if not isinstance(elev, (str, int, float)):
            return False

        elev = valueToFloat(elev)
        if elev is None:
            return False

        conn = Connection(self.host)

        setElev = ':Sev{sign}{0:06.1f}#'.format(abs(elev),
                                                sign='+' if elev > 0 else '-')

        commandString = setElev
        suc, response, numberOfChunks = conn.communicate(commandString)
        if not suc:
            return False

        if '0' in response[0]:
            return False

        return True

    def startTracking(self):
        """
        startTracking sends the start command to the mount. the command returns nothing.
        it is necessary to make that direct to unpark first, than start tracking

        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':PO#:AP#')
        return suc

    def stopTracking(self):
        """
        stopTracking sends the start command to the mount. the command returns nothing.

        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':RT9#')
        return suc

    def park(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':hP#')
        return suc

    def unpark(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':PO#')
        return suc

    def parkOnActualPosition(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':PiP#')
        if not suc:
            return False

        if '0' in response[0]:
            return False

        return True

    def stop(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':STOP#')
        return suc

    def flip(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':FLIP#')
        if not suc:
            return False

        if response[0] != '1':
            return False

        return True

    def moveNorth(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':PO#:Mn#')
        if not suc:
            return False

        return True

    def moveEast(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':PO#:Me#')
        if not suc:
            return False

        return True

    def moveSouth(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':PO#:Ms#')
        if not suc:
            return False

        return True

    def moveWest(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':PO#:Mw#')
        if not suc:
            return False

        return True

    def stopMoveNorth(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':Qn#')
        if not suc:
            return False

        return True

    def stopMoveAll(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':Q#')
        if not suc:
            return False

        return True

    def stopMoveEast(self):
        """
        stopMoveEast sends the flip command to the mount.

        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':Qe#')
        if not suc:
            return False
        return True

    def stopMoveSouth(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':Qs#')
        if not suc:
            return False

        return True

    def stopMoveWest(self):
        """
        :return:    success
        """
        conn = Connection(self.host)
        suc, response, numberOfChunks = conn.communicate(':Ww#')
        if not suc:
            return False

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages

# local imports
from base.ascomClass import AscomClass


class DomeAscom(AscomClass):
    """
    the class Dome inherits all information and handling of the Dome device. there will be
    some parameters who will define the slewing position of the dome relating to the
    mount.dome = DomeAscom(app=None)
    """

    __all__ = ['DomeAscom',
               ]

    CYCLE_DEVICE = 3000
    CYCLE_DATA = 1000

    def __init__(self, app=None, signals=None, data=None):
        super().__init__(app=app, data=data, threadPool=app.threadPool)

        self.signals = signals
        self.data = data

    def processPolledData(self):
        """
        :return: true for test purpose
        """
        azimuth = self.data.get('ABS_DOME_POSITION.DOME_ABSOLUTE_POSITION', 0)
        self.signals.azimuth.emit(azimuth)

        return True

    def workerPollData(self):
        """
        :return: true for test purpose
        """
        shutterStates = ['Open', 'Closed', 'Opening', 'Closing', 'Error']

        if not self.deviceConnected:
            return False

        azimuth = self.client.Azimuth
        self.dataEntry(azimuth, 'ABS_DOME_POSITION.DOME_ABSOLUTE_POSITION')
        self.signals.azimuth.emit(azimuth)
        self.dataEntry(self.client.Slewing, 'Slewing')
        self.dataEntry(self.client.CanSetAltitude, 'CanSetAltitude')
        self.dataEntry(self.client.CanSetAzimuth, 'CanSetAzimuth')
        self.dataEntry(self.client.CanSetShutter, 'CanSetShutter')

        state = self.client.ShutterStatus
        stateText = shutterStates[state]
        self.dataEntry(stateText, 'Status.Shutter')
        if state == 0:
            self.dataEntry(True,
                           'DOME_SHUTTER.SHUTTER_OPEN',
                           elementInv='DOME_SHUTTER.SHUTTER_CLOSED')
        elif state == 1:
            self.dataEntry(False,
                           'DOME_SHUTTER.SHUTTER_OPEN',
                           elementInv='DOME_SHUTTER.SHUTTER_CLOSED')
        else:
            self.data['DOME_SHUTTER.SHUTTER_OPEN'] = None
            self.data['DOME_SHUTTER.SHUTTER_CLOSED'] = None

        return True

    def slewToAltAz(self, altitude=0, azimuth=0):
        """
        :param altitude:
        :param azimuth:
        :return: success
        """
        if not self.deviceConnected:
            return False

        if self.data.get('CanSetAzimuth'):
            self.callMethodThreaded(self.client.SlewToAzimuth, azimuth)
        if self.data.get('CanSetAltitude'):
            self.callMethodThreaded(self.client.SlewToAltitude, altitude)
        return True

    def openShutter(self):
        """
        :return: success
        """
        if not self.deviceConnected:
            return False

        if self.data.get('CanSetShutter'):
            self.callMethodThreaded(self.client.OpenShutter)
        return True

    def closeShutter(self):
        """
        :return: success
        """
        if not self.deviceConnected:
            return False

        if self.data.get('CanSetShutter'):
            self.callMethodThreaded(self.client.CloseShutter)
        return True

    def abortSlew(self):
        """
        :return: success
        """
        if not self.deviceConnected:
            return False

        self.callMethodThreaded(self.client.AbortSlew)
        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import os

# external packages
import PyQt5
import numpy as np
from sgp4.exporter import export_tle

# local import
from base.tpool import Worker
from logic.databaseProcessing.dataWriter import DataWriter


class Satellite(object):
    """
    """

    def __init__(self):
        self.satellites = dict()
        self.satellite = None
        self.satOrbits = None
        self.listSatelliteNamesProxy = None
        self.satellitesRawTLE = {}
        self.databaseProcessing = DataWriter(self.app)
        self.installPath = ''

        self.satelliteSourceURLs = {
            'Active': 'http://www.celestrak.com/NORAD/elements/active.txt',
            'Space Stations': 'http://www.celestrak.com/NORAD/elements/stations.txt',
            '100 brightest': 'http://www.celestrak.com/NORAD/elements/visual.txt',
            'NOAA': 'http://www.celestrak.com/NORAD/elements/noaa.txt',
            'GEOS': 'http://www.celestrak.com/NORAD/elements/goes.txt',
            'Weather': 'http://www.celestrak.com/NORAD/elements/weather.txt',
            'Earth Resources': 'http://celestrak.com/NORAD/elements/resource.txt',
            'TDRSS TRacking and Data Relay': 'http://celestrak.com/NORAD/elements/tdrss.txt',
            'ARGOS': 'http://celestrak.com/NORAD/elements/argos.txt',
            'Amateur Radio': 'http://www.celestrak.com/NORAD/elements/amateur.txt',
            'Space & Earth Science': 'http://celestrak.com/NORAD/elements/science.txt',
            'Engineering': 'http://celestrak.com/NORAD/elements/engineering.txt',
            'Last 30 days launch': 'http://www.celestrak.com/NORAD/elements/tle-new.txt',
        }

        self.ui.progSatellitesFull.clicked.connect(self.progSatellitesFull)
        self.ui.progSatellitesFiltered.clicked.connect(self.progSatellitesFiltered)
        self.ui.listSatelliteNames.doubleClicked.connect(self.signalExtractSatelliteData)
        self.ui.startSatelliteTracking.clicked.connect(self.startTrack)
        self.ui.stopSatelliteTracking.clicked.connect(self.stopTrack)
        self.ui.satelliteSource.currentIndexChanged.connect(self.loadTLEDataFromSourceURLs)
        self.ui.filterSatellite.textChanged.connect(self.filterSatelliteNamesList)

        self.app.sendSatelliteData.connect(self.sendSatelliteData)
        self.app.mount.signals.calcTLEdone.connect(self.updateSatelliteTrackGui)
        self.app.mount.signals.getTLEdone.connect(self.getSatelliteDataFromDatabase)
        self.ui.isOnline.stateChanged.connect(self.loadTLEDataFromSourceURLs)

        self.app.update3s.connect(self.updateOrbit)

    def initConfig(self):
        """
        :return: True for test purpose
        """
        self.setupSatelliteSourceURLsDropDown()
        if not self.app.automation:
            self.installPath = self.app.mwGlob['dataDir']

        elif self.app.automation.installPath:
            self.installPath = self.app.automation.installPath

        else:
            self.installPath = self.app.mwGlob['dataDir']
        return True

    def setupSatelliteSourceURLsDropDown(self):
        """
        setupSatelliteSourceURLsDropDown handles the dropdown list for the
        satellite data online sources. therefore we add the necessary entries to
        populate the list.

        :return: success for test
        """
        self.ui.satelliteSource.clear()
        self.ui.satelliteSource.setView(PyQt5.QtWidgets.QListView())
        for name in self.satelliteSourceURLs.keys():
            self.ui.satelliteSource.addItem(name)
        return True

    def filterSatelliteNamesList(self):
        """
        :return: true for test purpose
        """
        listSat = self.ui.listSatelliteNames
        filterStr = self.ui.filterSatellite.text()

        for row in range(listSat.model().rowCount()):
            isFound = filterStr.lower() in listSat.model().index(row).data().lower()
            isVisible = isFound or not filterStr

            if isVisible:
                listSat.setRowHidden(row, False)

            else:
                listSat.setRowHidden(row, True)
        return True

    def setupSatelliteNameList(self):
        """
        setupSatelliteNameList clears the list view of satellite names deriving
        from the selected source file on disk. after that it populated the list
        with actual data.

        :return: success for test
        """
        self.ui.listSatelliteNames.clear()
        for name, _ in self.satellites.items():
            if not isinstance(name, str):
                continue
            entryName = f'{self.satellites[name].model.satnum:6d}: {name}'
            self.ui.listSatelliteNames.addItem(entryName)
        self.ui.listSatelliteNames.sortItems()
        self.ui.listSatelliteNames.update()
        self.filterSatelliteNamesList()
        return True

    def loadTLEDataFromSourceURLsWorker(self, source='', isOnline=False):
        """
        loadTLEDataFromSourceURLsWorker selects from a drop down list of
        possible satellite data sources on the web and once selected downloads
        the data. depending of the setting of reload is true setting, it takes an
        already loaded file from local disk. after loading or opening the source
        file, it updates the satellite list in the list view widget for the
        selection of satellites.

        :return: success
        """
        if not source:
            return False

        fileName = os.path.basename(source)
        dirPath = self.app.mwGlob['dataDir']
        filePath = f'{dirPath}/{fileName}'

        satellites = self.app.mount.obsSite.loader.tle_file(source, reload=isOnline)

        self.satellites = {sat.name: sat for sat in satellites}

        if not os.path.isfile(filePath):
            return False
        return True

    def loadTLEDataFromSourceURLs(self):
        """
        loadTLEDataFromSourceURLs selects from a drop down list of possible
        satellite data sources on the web and once selected downloads the data.
        depending of the setting of reload is true setting, it takes an already
        loaded file from local disk. after loading or opening the source file,
        it updates the satellite list in the list view widget for the selection
        of satellites.

        :return: success
        """
        key = self.ui.satelliteSource.currentText()
        if key not in self.satelliteSourceURLs:
            return False

        source = self.satelliteSourceURLs[key]
        isOnline = self.ui.isOnline.isChecked()
        worker = Worker(self.loadTLEDataFromSourceURLsWorker,
                        source=source,
                        isOnline=isOnline)
        worker.signals.finished.connect(self.setupSatelliteNameList)
        self.threadPool.start(worker)

        return True

    def updateOrbit(self):
        """
        updateOrbit calculates the actual satellite orbits, sub point etc. and
        updates the data in the gui. in addition when satellite window is open
        it signals this update data as well for matplotlib drawings in satellite
        window. this method is called cyclic every 3 seconds for updates

        :return: success
        """
        if self.satellite is None:
            self.ui.startSatelliteTracking.setEnabled(False)
            return False

        # check if calculation is necessary to optimize cpu time
        # get index for satellite tab and check if it's visible. i
        # f not, no calculation
        tabWidget = self.ui.mainTabWidget.findChild(PyQt5.QtWidgets.QWidget, 'Satellite')
        tabIndex = self.ui.mainTabWidget.indexOf(tabWidget)
        satTabVisible = self.ui.mainTabWidget.currentIndex() == tabIndex

        winObj = self.app.uiWindows['showSatelliteW']

        if not winObj:
            return False

        # if nothing is visible, nothing to update !
        if not winObj.get('classObj') and not satTabVisible:
            return False

        # now calculating the satellite data
        now = self.app.mount.obsSite.ts.now()
        observe = self.satellite.at(now)

        subpoint = observe.subpoint()
        difference = self.satellite - self.app.mount.obsSite.location

        ra, dec, _ = difference.at(now).radec()
        ra = ra.hours
        dec = dec.degrees

        self.ui.satRa.setText(f'{ra:3.2f}')
        self.ui.satDec.setText(f'{dec:3.2f}')

        lat = subpoint.latitude.degrees
        lon = subpoint.longitude.degrees
        self.ui.satLatitude.setText(f'{lat:3.2f}')
        self.ui.satLongitude.setText(f'{lon:3.2f}')

        altaz = difference.at(now).altaz()

        alt, az, _ = altaz
        alt = alt.degrees
        az = az.degrees

        self.ui.satAltitude.setText(f'{alt:3.2f}')
        self.ui.satAzimuth.setText(f'{az:3.2f}')

        if not winObj.get('classObj'):
            return True

        winObj['classObj'].signals.update.emit(observe, subpoint, altaz)

        return True

    def programTLEDataToMount(self):
        """
        programTLEDataToMount get the satellite data and programs this TLE data
        into the mount. after programming the parameters it forces the mount to
        calculate the satellite orbits immediately

        :return: success
        """
        if not self.app.mount.mountUp:
            self.app.message.emit('Mount is not online', 2)
            return False

        satellite = self.app.mount.satellite
        if satellite.tleParams.name == self.satellite.name:
            self.app.message.emit(f'Actual satellite is  [{self.satellite.name}]', 0)

        else:
            self.app.message.emit(f'Programming [{self.satellite.name}] to mount', 0)
            line0 = self.satellite.name
            line1, line2 = export_tle(self.satellite.model)
            suc = satellite.setTLE(line0=line0,
                                   line1=line1,
                                   line2=line2)
            if not suc:
                self.app.message.emit('Error program TLE', 2)
                return False
        return True

    def calcOrbitFromTLEInMount(self):
        """
        :return: success
        """
        if self.satellite is None:
            self.ui.startSatelliteTracking.setEnabled(False)
            self.ui.stopSatelliteTracking.setEnabled(False)
            return False

        self.app.mount.calcTLE()
        return True

    def showRises(self):
        """
        showRises calculated the next three satellite passes for the
        presentation in the gui. the times shown might differ from the
        calculation of the mount as we dont know, how the mount calculates is
        timings.

        :return: True for test purpose
        """
        minAlt = self.app.mount.setting.horizonLimitLow
        if minAlt is None:
            minAlt = 0

        loc = self.app.mount.obsSite.location
        obs = self.app.mount.obsSite

        orbitCycleTime = np.pi / self.satellite.model.no_kozai / 12 / 60

        t0 = obs.ts.tt_jd(obs.timeJD.tt - orbitCycleTime)
        t1 = obs.ts.tt_jd(obs.timeJD.tt + 3)

        t, events = self.satellite.find_events(loc, t0, t1, altitude_degrees=minAlt)

        passUI = {
            0:
                {'rise': self.ui.satTransitStartUTC_1,
                 'settle': self.ui.satTransitEndUTC_1,
                 },
            1:
                {'rise': self.ui.satTransitStartUTC_2,
                 'settle': self.ui.satTransitEndUTC_2,
                 },
            2:
                {'rise': self.ui.satTransitStartUTC_3,
                 'settle': self.ui.satTransitEndUTC_3,
                 },
        }

        fString = "%Y-%m-%d  %H:%M"

        # collecting the events
        index = 0
        satOrbits = dict()
        for ti, event in zip(t, events):
            if event == 0:
                satOrbits[index] = {'rise': ti}

            elif event == 1:
                continue

            elif event == 2:
                if index not in satOrbits:
                    continue

                satOrbits[index]['settle'] = ti

                if ti.tt < obs.timeJD.tt:
                    continue

                index += 1

        for satOrbit in satOrbits:
            if satOrbit > 2:
                break
            timeRise = satOrbits[satOrbit]['rise'].utc_strftime(fString)
            timeSettle = satOrbits[satOrbit]['settle'].utc_strftime(fString)
            passUI[satOrbit]['rise'].setText(f'{timeRise}')
            passUI[satOrbit]['settle'].setText(f'{timeSettle}')

        while index < 3:
            passUI[index]['rise'].setText('-')
            passUI[index]['settle'].setText('-')
            index += 1

        return satOrbits

    def extractSatelliteData(self, satName=''):
        """
        extractSatelliteData is called when a satellite is selected via mouse
        click in the list menu. it collects the data and writes basic stuff to
        the gui. depending on the age of the satellite data is colors the frame

        :param satName: additional parameter for calling this method
        :return: success
        """
        if satName not in self.satellites:
            return False

        index = self.findIndexValue(self.ui.listSatelliteNames, satName,
                                    relaxed=True)
        item = self.ui.listSatelliteNames.item(index)

        if item is None:
            return False

        item.setSelected(True)

        # making the entry visible (and scroll the list if necessary)
        position = PyQt5.QtWidgets.QAbstractItemView.EnsureVisible
        self.ui.listSatelliteNames.scrollToItem(item, position)
        self.satellite = self.satellites[satName]

        # now we prepare the selection of the data in the gui
        self.ui.satelliteName.setText(self.satellite.name)
        epochText = self.satellite.epoch.utc_strftime('%Y-%m-%d, %H:%M')
        self.ui.satelliteEpoch.setText(epochText)

        # the epoch should be not too old
        now = self.app.mount.obsSite.ts.now()
        days = now - self.satellite.epoch
        self.ui.satelliteDataAge.setText(f'{days:2.2f}')

        if days > 10:
            self.changeStyleDynamic(self.ui.satelliteDataAge, 'color', 'red')

        elif 3 < days < 10:
            self.changeStyleDynamic(self.ui.satelliteDataAge, 'color', 'yellow')

        else:
            self.changeStyleDynamic(self.ui.satelliteDataAge, 'color', '')

        # filling up the satellite data
        self.ui.satelliteNumber.setText(f'{self.satellite.model.satnum:5d}')
        self.ui.stopSatelliteTracking.setEnabled(False)
        self.ui.startSatelliteTracking.setEnabled(False)
        self.ui.satTransitStartUTC.setText('-')
        self.ui.satTransitEndUTC.setText('-')
        self.ui.satNeedFlip.setText('-')

        self.updateOrbit()
        self.programTLEDataToMount()
        self.calcOrbitFromTLEInMount()
        self.satOrbits = self.showRises()

        winObj = self.app.uiWindows['showSatelliteW']

        if not winObj['classObj']:
            return False

        winObj['classObj'].signals.show.emit(self.satellite, self.satOrbits)
        return True

    def sendSatelliteData(self):
        """
        :return:
        """
        if not self.satellite or not self.satOrbits:
            return False

        winObj = self.app.uiWindows['showSatelliteW']

        if not winObj['classObj']:
            return False

        winObj['classObj'].signals.show.emit(self.satellite, self.satOrbits)
        return True

    def signalExtractSatelliteData(self):
        """
        :return: True for test purpose
        """
        satName = self.ui.listSatelliteNames.currentItem().text()[8:]
        self.extractSatelliteData(satName=satName)
        return True

    def getSatelliteDataFromDatabase(self, tleParams=None):
        """
        getSatelliteDataFromDatabase gets called, when the TLE setup is read
        from the mount. we use the name to retrieve the data from the
        "active.txt" database to be able to work with external database. it calls
        extraction method for getting the specific satellite data read and stored.

        :param tleParams:
        :return: True for test purpose
        """
        if not tleParams:
            return False

        self.extractSatelliteData(satName=tleParams.name)
        return True

    def updateSatelliteTrackGui(self, tleParams=None):
        """
        updateSatelliteTrackGui is called, when the mount has finished its
        calculations based on programmed TLE data. It writes the data to the gui
        and enables the start track button.

        :return: success for test purpose
        """
        if not tleParams:
            return False

        if tleParams.jdStart is not None:
            time = self.app.mount.obsSite.ts.tt_jd(tleParams.jdStart)
            self.ui.satTransitStartUTC.setText(time.utc_strftime('%Y-%m-%d  %H:%M:%S'))

        else:
            self.ui.satTransitStartUTC.setText('No transit')

        if tleParams.jdEnd is not None:
            time = self.app.mount.obsSite.ts.tt_jd(tleParams.jdEnd)
            self.ui.satTransitEndUTC.setText(time.utc_strftime('%Y-%m-%d  %H:%M:%S'))

        else:
            self.ui.satTransitEndUTC.setText('No transit')

        if tleParams.flip:
            self.ui.satNeedFlip.setText('YES')

        else:
            self.ui.satNeedFlip.setText('NO')

        if tleParams.message is not None:
            self.app.message.emit(tleParams.message, 0)

        if tleParams.altitude is not None:
            self.ui.stopSatelliteTracking.setEnabled(True)
            self.ui.startSatelliteTracking.setEnabled(True)
        else:
            self.ui.stopSatelliteTracking.setEnabled(False)
            self.ui.startSatelliteTracking.setEnabled(False)

        return True

    def startTrack(self):
        """
        :return: success
        """
        if not self.app.mount.mountUp:
            self.app.message.emit('Mount is not online', 2)
            return False

        # if mount is currently parked, we unpark it
        if self.app.mount.obsSite.status == 5:
            suc = self.app.mount.obsSite.unpark()
            if suc:
                self.app.message.emit('Mount unparked', 0)

            else:
                self.app.message.emit('Cannot unpark mount', 2)

        suc, message = self.app.mount.satellite.slewTLE()
        if not suc:
            self.app.message.emit(message, 2)
            return False

        self.app.message.emit(message, 0)
        return True

    def stopTrack(self):
        """
        :return: success
        """
        if not self.app.mount.mountUp:
            self.app.message.emit('Mount is not online', 2)
            return False

        suc = self.app.mount.obsSite.stopTracking()
        if not suc:
            self.app.message.emit('Cannot stop tracking', 2)
        else:
            self.app.message.emit('Stopped tracking', 0)
        return suc

    def progSatellitesFiltered(self):
        """
        :return: success
        """
        source = self.ui.satelliteSource.currentText()
        text = f'Should filtered database\n\n[{source}]\n\nbe programmed to mount ?'
        suc = self.messageDialog(self, 'Program with QCI Updater', text)

        if not suc:
            return False

        self.app.message.emit(f'Program database:    [{source}]', 1)
        self.app.message.emit('Exporting MPC data', 0)

        filterStr = self.ui.filterSatellite.text().lower()

        filtered = dict()
        for name, _ in self.satellites.items():
            if not isinstance(name, str):
                continue

            text = f'{self.satellites[name].model.satnum:6d}: {name}'
            if filterStr.lower() not in text.lower():
                continue

            filtered[name] = self.satellites[name]

        suc = self.databaseProcessing.writeSatelliteTLE(filtered,
                                                        self.installPath)

        if not suc:
            self.app.message.emit('Data could not be exported - stopping', 2)
            return False

        if not self.app.automation:
            self.app.message.emit('Not running windows, no updater available', 2)
            return False

        self.app.message.emit('Uploading TLE data to mount', 0)
        suc = self.app.automation.uploadTLEData()

        if not suc:
            self.app.message.emit('Uploading error', 2)

        self.app.message.emit('Programming success', 1)

        return suc

    def progSatellitesFull(self):
        """
        :return: success
        """
        source = self.ui.satelliteSource.currentText()
        text = f'Should full database\n\n[{source}]\n\nbe programmed to mount ?'
        suc = self.messageDialog(self, 'Program with QCI Updater', text)

        if not suc:
            return False

        self.app.message.emit(f'Program database:    [{source}]', 1)
        self.app.message.emit('Exporting TLE data', 0)

        suc = self.databaseProcessing.writeSatelliteTLE(self.satellites,
                                                        self.installPath)

        if not suc:
            self.app.message.emit('Data could not be exported - stopping', 2)
            return False

        if not self.app.automation:
            self.app.message.emit('Not running windows, no updater available', 2)
            return False

        self.app.message.emit('Uploading TLE data to mount', 0)
        suc = self.app.automation.uploadTLEData()

        if not suc:
            self.app.message.emit('Uploading error', 2)

        self.app.message.emit('Programming success', 1)
        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages

# local imports
from base.indiClass import IndiClass


class WeatherUPBIndi(IndiClass):
    """
    the class PegasusUPBIndi inherits all information and handling of the PegasusUPB device

        >>> power = WeatherUPBIndi(app=None)
    """

    __all__ = ['WeatherUPBIndi',
               ]

    # update rate to 1000 milli seconds for setting indi server
    UPDATE_RATE = 1000

    def __init__(self, app=None, signals=None, data=None):
        super().__init__(app=app, data=data, threadPool=app.threadPool)

        self.signals = signals
        self.data = data

    def setUpdateConfig(self, deviceName):
        """
        _setUpdateRate corrects the update rate of weather devices to get an defined
        setting regardless, what is setup in server side.

        :param deviceName:
        :return: success
        """

        if deviceName != self.deviceName:
            return False

        if self.device is None:
            return False

        # setting polling updates in driver

        update = self.device.getNumber('POLLING')

        if 'PERIOD' not in update:
            return False

        if update.get('PERIOD', 0) == self.UPDATE_RATE:
            return True

        update['PERIOD'] = self.UPDATE_RATE
        suc = self.client.sendNewNumber(deviceName=deviceName,
                                        propertyName='POLLING',
                                        elements=update,
                                        )

        return suc



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages

# local imports
from base.indiClass import IndiClass


class CoverIndi(IndiClass):
    """
    the class FlipFlatIndi inherits all information and handling of the FlipFlat device

        >>> f = CoverIndi(app=None)
    """

    __all__ = ['CoverIndi',
               ]

    # update rate to 1 seconds for setting indi server
    UPDATE_RATE = 1

    def __init__(self, app=None, signals=None, data=None):
        super().__init__(app=app, threadPool=app.threadPool)

        self.signals = signals
        self.data = data

    def setUpdateConfig(self, deviceName):
        """
        _setUpdateRate corrects the update rate of weather devices to get an defined
        setting regardless, what is setup in server side.

        :param deviceName:
        :return: success
        """
        if deviceName != self.deviceName:
            return False

        if self.device is None:
            return False

        update = self.device.getNumber('PERIOD_MS')

        if 'PERIOD' not in update:
            return False

        if update.get('PERIOD', 0) == self.UPDATE_RATE:
            return True

        update['PERIOD'] = self.UPDATE_RATE
        suc = self.client.sendNewNumber(deviceName=deviceName,
                                        propertyName='PERIOD_MS',
                                        elements=update)
        return suc

    def updateText(self, deviceName, propertyName):
        """
        :param deviceName:
        :param propertyName:
        :return:
        """
        if not super().updateText(deviceName, propertyName):
            return False

        for element, value in self.device.getText(propertyName).items():
            if element == 'Cover':
                value = value.strip().upper()
                if value == 'OPEN':
                    self.data['CAP_PARK.UNPARK'] = 'On'
                    self.data['CAP_PARK.PARK'] = 'Off'

                elif value == 'CLOSED':
                    self.data['CAP_PARK.UNPARK'] = 'Off'
                    self.data['CAP_PARK.PARK'] = 'On'

                else:
                    self.data['CAP_PARK.UNPARK'] = None
                    self.data['CAP_PARK.PARK'] = None

        return True

    def closeCover(self):
        """
        :return: success
        """
        if self.device is None:
            return False

        cover = self.device.getSwitch('CAP_PARK')

        if 'PARK' not in cover:
            return False

        cover['UNPARK'] = 'Off'
        cover['PARK'] = 'On'

        suc = self.client.sendNewSwitch(deviceName=self.deviceName,
                                        propertyName='CAP_PARK',
                                        elements=cover,
                                        )
        return suc

    def openCover(self):
        """
        :return: success
        """
        if self.device is None:
            return False

        cover = self.device.getSwitch('CAP_PARK')

        if 'UNPARK' not in cover:
            return False

        cover['UNPARK'] = 'On'
        cover['PARK'] = 'Off'

        suc = self.client.sendNewSwitch(deviceName=self.deviceName,
                                        propertyName='CAP_PARK',
                                        elements=cover,
                                        )
        return suc

    @staticmethod
    def haltCover():
        """
        :return: success
        """
        return False



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import logging

# external packages
import skyfield.api

# local imports
from mountcontrol.convert import valueToAngle
from mountcontrol.convert import stringToAngle


class AlignStar(object):
    """
    The class AlignStar inherits all information and handling of the coordinates of
    a pointing direction of the nominal mount coordinates and the plate solved
    coordinates, which were derived from a taken image at the scope
    the coordinates both are in JNow topocentric

        >>> settings = AlignStar(
        >>>                     )

    coordinates could be from type skyfield.api.Star or just a tuple of (ha, dec) where
    the format should be float or the 10micron string format.

    """

    __all__ = ['AlignStar',
               ]

    log = logging.getLogger(__name__)

    def __init__(self,
                 mCoord=None,
                 pierside=None,
                 sCoord=None,
                 sidereal=None,
                 ):

        self.mCoord = mCoord
        self.pierside = pierside
        self.sCoord = sCoord
        self.sidereal = sidereal

    @property
    def mCoord(self):
        return self._mCoord

    @mCoord.setter
    def mCoord(self, value):
        if isinstance(value, skyfield.api.Star):
            self._mCoord = value
            return

        if not isinstance(value, (tuple, list)):
            self._mCoord = None
            return

        if len(value) != 2:
            self._mCoord = None
            return

        ra, dec = value
        if isinstance(ra, str):
            ra = stringToAngle(ra, preference='hours')
        if isinstance(ra, float):
            ra = valueToAngle(ra, preference='hours')
        if isinstance(dec, str):
            dec = stringToAngle(dec)
        if isinstance(dec, float):
            dec = valueToAngle(dec)

        if not ra or not dec:
            self._mCoord = None
            self.log.warning('Malformed value: {0}'.format(value))
            return

        self._mCoord = skyfield.api.Star(ra=ra,
                                         dec=dec)

    @property
    def pierside(self):
        return self._pierside

    @pierside.setter
    def pierside(self, value):
        if value in ['E', 'W', 'e', 'w']:
            value = value.capitalize()
            self._pierside = value

        else:
            self._pierside = None
            self.log.warning('Malformed value: {0}'.format(value))

    @property
    def sCoord(self):
        return self._sCoord

    @sCoord.setter
    def sCoord(self, value):
        if isinstance(value, skyfield.api.Star):
            self._sCoord = value
            return

        if not isinstance(value, (tuple, list)):
            self._sCoord = None
            return

        if len(value) != 2:
            self._sCoord = None
            return

        ra, dec = value
        if isinstance(ra, str):
            ra = stringToAngle(ra, preference='hours')
        if isinstance(ra, float):
            ra = valueToAngle(ra, preference='hours')
        if isinstance(dec, str):
            dec = stringToAngle(dec)
        if isinstance(dec, float):
            dec = valueToAngle(dec)

        if not ra or not dec:
            self._sCoord = None
            self.log.warning('Malformed value: {0}'.format(value))
            return

        self._sCoord = skyfield.api.Star(ra=ra,
                                         dec=dec)

    @property
    def sidereal(self):
        return self._sidereal

    @sidereal.setter
    def sidereal(self, value):
        if isinstance(value, str):
            self._sidereal = stringToAngle(value, preference='hours')

        elif isinstance(value, float):
            self._sidereal = valueToAngle(value, preference='hours')

        elif isinstance(value, skyfield.api.Angle):
            self._sidereal = value

        else:
            self._sidereal = None



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
#
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages
import PyQt5.QtCore
import PyQt5.QtWidgets
from deepdiff import DeepDiff

# local import
from gui.extWindows.devicePopupW import DevicePopup


class SettDevice:
    """
    devices types in self.drivers are name related to ascom definitions

    architecture:
    - all properties are stored in the config dict as main source.
    - when starting, all gui elements will be populated based on the entries of config
    - all drivers were initialised with the content of config dict
    - if we setup a new device, data of device is gathered for the popup from config
    - when closing the popup, result data will be stored in config
    - if there is no default data for a driver in config dict, it will be retrieved from the
      driver

    sequence standard:
        loading config dict
        load driver default setup from driver if not present in config
        initialize gui
        initialize driver
        start drivers

    sequence popup:
        initialize popup data
        call popup modal
        popup close
        if cancel -> finished
        store data in config
        stop changed driver
        start new driver

    sequence dropdown:
        search driver
        if no driver -> finished
        stop changed driver
        if driver = "device disabled" -> finished
        start new driver
    """

    def __init__(self, app=None, ui=None, clickable=None):
        if app:
            self.app = app
            self.ui = ui
            self.clickable = clickable

        self.popupUi = None
        self.drivers = {
            'dome': {
                'uiDropDown': self.ui.domeDevice,
                'uiSetup': self.ui.domeSetup,
                'class': self.app.dome,
                'deviceType': 'dome',
            },
            'cover': {
                'uiDropDown': self.ui.coverDevice,
                'uiSetup': self.ui.coverSetup,
                'class': self.app.cover,
                'deviceType': 'covercalibrator',
            },
            'camera': {
                'uiDropDown': self.ui.cameraDevice,
                'uiSetup': self.ui.cameraSetup,
                'class': self.app.camera,
                'deviceType': 'camera',
            },
            'filter': {
                'uiDropDown': self.ui.filterDevice,
                'uiSetup': self.ui.filterSetup,
                'class': self.app.filter,
                'deviceType': 'filterwheel',
            },
            'focuser': {
                'uiDropDown': self.ui.focuserDevice,
                'uiSetup': self.ui.focuserSetup,
                'class': self.app.focuser,
                'deviceType': 'focuser',
            },
            'sensorWeather': {
                'uiDropDown': self.ui.sensorWeatherDevice,
                'uiSetup': self.ui.sensorWeatherSetup,
                'class': self.app.sensorWeather,
                'deviceType': 'observingconditions',
            },
            'directWeather': {
                'uiDropDown': self.ui.directWeatherDevice,
                'uiSetup': None,
                'class': self.app.directWeather,
                'deviceType': None,
            },
            'onlineWeather': {
                'uiDropDown': self.ui.onlineWeatherDevice,
                'uiSetup': self.ui.onlineWeatherSetup,
                'class': self.app.onlineWeather,
                'deviceType': None,
            },
            'skymeter': {
                'uiDropDown': self.ui.skymeterDevice,
                'uiSetup': self.ui.skymeterSetup,
                'class': self.app.skymeter,
                'deviceType': 'observingconditions',
            },
            'powerWeather': {
                'uiDropDown': self.ui.powerWeatherDevice,
                'uiSetup': self.ui.powerWeatherSetup,
                'class': self.app.powerWeather,
                'deviceType': 'observingconditions',
            },
            'telescope': {
                'uiDropDown': self.ui.telescopeDevice,
                'uiSetup': self.ui.telescopeSetup,
                'class': self.app.telescope,
                'deviceType': 'telescope',
            },
            'power': {
                'uiDropDown': self.ui.powerDevice,
                'uiSetup': self.ui.powerSetup,
                'class': self.app.power,
                'deviceType': 'power',
            },
            'relay': {
                'uiDropDown': self.ui.relayDevice,
                'uiSetup': self.ui.relaySetup,
                'class': self.app.relay,
                'deviceType': None,
            },
            'astrometry': {
                'uiDropDown': self.ui.astrometryDevice,
                'uiSetup': self.ui.astrometrySetup,
                'class': self.app.astrometry,
                'deviceType': 'astrometry',
            },
            'remote': {
                'uiDropDown': self.ui.remoteDevice,
                'uiSetup': None,
                'class': self.app.remote,
                'deviceType': None,
            },
            'measure': {
                'uiDropDown': self.ui.measureDevice,
                'uiSetup': None,
                'class': self.app.measure,
                'deviceType': None,
            },
        }

        self.driversData = {}

        for driver in self.drivers:
            self.drivers[driver]['uiDropDown'].activated.connect(self.dispatchDriverDropdown)
            if self.drivers[driver]['uiSetup'] is not None:
                ui = self.drivers[driver]['uiSetup']
                ui.clicked.connect(self.dispatchPopup)

            if hasattr(self.drivers[driver]['class'], 'signals'):
                signals = self.drivers[driver]['class'].signals
                signals.serverDisconnected.connect(self.serverDisconnected)
                signals.deviceConnected.connect(self.deviceConnected)
                signals.deviceDisconnected.connect(self.deviceDisconnected)

        self.ui.ascomConnect.clicked.connect(self.manualStartAllAscomDrivers)
        self.ui.ascomDisconnect.clicked.connect(self.manualStopAllAscomDrivers)

    def checkStructureDriversData(self, driver, config):
        """
        checkStructureDriversData
        :param driver:
        :param config:
        :return:
        """

        defaultConfig = self.drivers[driver]['class'].defaultConfig
        res = DeepDiff(defaultConfig, config['driversData'][driver])

        if 'dictionary_item_added' in res or 'dictionary_item_removed' in res:
            config['driversData'][driver] = defaultConfig
            self.log.info(f'Config for {[driver]} updated to default')
            return False

        return True

    def setDefaultData(self, driver, config):
        """

        :param driver:
        :param config:
        :return:
        """

        config['driversData'][driver] = {}
        defaultConfig = self.drivers[driver]['class'].defaultConfig
        config['driversData'][driver].update(defaultConfig)

        return True

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict. if some drivers configuration
        are missing, the default values are loaded from the drivers. after that we have a
        fully cleaned config dict and we can proceed initializing the gui and the drivers

        :return: True for test purpose
        """

        config = self.app.config['mainW']

        if 'driversData' not in config:
            config['driversData'] = {}

        for driver in self.drivers:
            if driver in config['driversData']:
                self.checkStructureDriversData(driver, config)
            else:
                self.setDefaultData(driver, config)

        self.driversData.update(config.get('driversData', {}))
        self.ui.checkASCOMAutoConnect.setChecked(config.get('checkASCOMAutoConnect', False))

        self.setupDeviceGui()
        self.startDrivers()

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """

        config = self.app.config['mainW']

        config['driversData'] = self.driversData
        config['checkASCOMAutoConnect'] = self.ui.checkASCOMAutoConnect.isChecked()

        return True

    def setupDeviceGui(self):
        """
        setupDeviceGui handles the dropdown lists for all devices possible. it reads the
        information out of the config dict and populated the entries where necessary

        :return: success for test
        """

        dropDowns = list(self.drivers[driver]['uiDropDown'] for driver in self.drivers)
        for dropDown in dropDowns:
            dropDown.clear()
            dropDown.setView(PyQt5.QtWidgets.QListView())
            dropDown.addItem('device disabled')

        for driver in self.driversData:
            frameworks = self.driversData[driver]['frameworks']

            if driver not in self.drivers:
                self.log.critical(f'Missing driver: [{driver}]')
                continue

            for fw in frameworks:
                name = frameworks[fw]['deviceName']
                itemText = f'{fw} - {name}'
                self.drivers[driver]['uiDropDown'].addItem(itemText)

            framework = self.driversData[driver]['framework']
            index = self.findIndexValue(self.drivers[driver]['uiDropDown'], framework)
            self.drivers[driver]['uiDropDown'].setCurrentIndex(index)

        return True

    def processPopupResults(self):
        """
        processPopupResults takes sets the actual drop down in the device settings lists to
        the choice of the popup window. after that it starts the driver again.

        :return: success if new device could be selected
        """

        self.popupUi.ui.ok.clicked.disconnect(self.processPopupResults)
        driver = self.popupUi.returnValues.get('driver')

        if not driver:
            return False

        if self.popupUi.returnValues.get('indiCopyConfig', False):
            self.copyConfig(driver=driver, framework='indi')

        if self.popupUi.returnValues.get('alpacaCopyConfig', False):
            self.copyConfig(driver=driver, framework='alpaca')

        selectedFramework = self.driversData[driver]['framework']
        index = self.findIndexValue(self.drivers[driver]['uiDropDown'], selectedFramework)
        name = self.driversData[driver]['frameworks'][selectedFramework]['deviceName']

        if not name:
            return False

        itemText = f'{selectedFramework} - {name}'

        self.drivers[driver]['uiDropDown'].setCurrentIndex(index)
        self.drivers[driver]['uiDropDown'].setItemText(index, itemText)
        self.drivers[driver]['uiDropDown'].update()

        self.stopDriver(driver=driver)
        self.startDriver(driver=driver)

        return True

    def copyConfig(self, driver='', framework=''):
        """
        copyConfig transfers all information of the actual driver to all other drivers of
        the same framework. if done so, all other drivers running on the same framework have
        to be stopped, copied parameters, initialized and started, too.

        :param driver:
        :param framework:
        :return: True for test purpose
        """

        for driverLoop in self.drivers:
            if driverLoop == driver:
                # not copy on the same driver
                continue

            driverClass = self.drivers[driverLoop]['class']

            if driverClass.framework == framework:
                self.stopDriver(driver=driver)

            if driverLoop not in self.driversData:
                continue

            if framework not in self.driversData[driverLoop]['frameworks']:
                continue

            for param in self.driversData[driverLoop]['frameworks'][framework]:
                if param in ['deviceList', 'deviceName']:
                    # should be not copied
                    continue

                source = self.driversData[driver]['frameworks'][framework][param]
                self.driversData[driverLoop]['frameworks'][framework][param] = source

            if driverClass.framework == framework:
                self.startDriver(driver=driver)

        return True

    def callPopup(self, driver):
        """
        callPopup prepares the data and calls and processes the returned data for the given
        driver.
        if copy flags for some frameworks are given, the properties of the actual driver
        will be copied to all other drivers with the same framework.

        :param driver:
        :return: True for test purpose
        """

        data = self.driversData[driver]
        deviceType = self.drivers[driver]['deviceType']

        self.popupUi = DevicePopup(self,
                                   app=self.app,
                                   driver=driver,
                                   deviceType=deviceType,
                                   data=data)

        self.popupUi.ui.ok.clicked.connect(self.processPopupResults)

        return True

    def dispatchPopup(self):
        """

        :return: True for test purpose
        """

        sender = self.sender()
        driver = self.returnDriver(sender, self.drivers, addKey='uiSetup')

        if driver:
            self.callPopup(driver=driver)

        return True

    def stopDriver(self, driver=None):
        """
        stopDriver stops the named driver.

        :param driver:
        :return: returns status if we are finished
        """

        if not driver:
            return False

        self.deviceStat[driver] = None

        framework = self.drivers[driver]['class'].framework

        if not framework:
            return False

        driverClass = self.drivers[driver]['class']
        isRunning = driverClass.run[framework].deviceName != ''

        if isRunning:
            driverClass.stopCommunication()
            driverClass.data.clear()
            driverClass.run[framework].deviceName = ''
            self.app.message.emit(f'Disabled device:     [{driver}]', 0)

        self.drivers[driver]['uiDropDown'].setStyleSheet(self.BACK_NORM)
        self.deviceStat[driver] = None

        return True

    def stopDrivers(self):
        """
        stopDrivers runs through all drivers and stops them

        :return: True for test purpose
        """
        for driver in self.drivers:
            self.stopDriver(driver=driver)

        return True

    def configDriver(self, driver=None):
        """
        configDriver configures a driver if a framework has been set. otherwise does nothing

        :param driver:
        :return: success of config
        """

        if not driver:
            return False

        self.deviceStat[driver] = False

        framework = self.driversData[driver]['framework']
        if not framework:
            return False

        frameworkConfig = self.driversData[driver]['frameworks'][framework]
        driverClass = self.drivers[driver]['class'].run[framework]

        for attribute in frameworkConfig:
            setattr(driverClass, attribute, frameworkConfig[attribute])

        return True

    def startDriver(self, driver=None, autoStart=True):
        """
        startDriver checks if a framework has been set and starts the driver with its
        startCommunication method. Normally the driver would report it's connection,
        but for internal driver this has to be done separately.

        as the configuration is stored in the config, start also stores the selected
        framework in the framework attribute of the driver's class. this is needed as when
        stopping the driver the config dict already has the new framework set and we have to
        remember it.

        :param driver:
        :param autoStart:
        :return: success if a driver has ben started
        """

        if not driver:
            return False

        data = self.driversData[driver]
        framework = data['framework']

        if not framework:
            return False

        loadConfig = data.get('loadConfig', False)
        driverClass = self.drivers[driver]['class']

        driverClass.framework = framework
        isInternal = framework == 'internal'

        if isInternal:
            self.drivers[driver]['uiDropDown'].setStyleSheet(self.BACK_GREEN)

        self.configDriver(driver=driver)

        if autoStart:
            driverClass.startCommunication(loadConfig=loadConfig)

        self.app.message.emit(f'Enabled device:      [{driver}]', 0)

        return True

    def startDrivers(self):
        """
        startDrivers starts all drivers
        and managing the boot / shutdown.

        :return: true for test purpose
        """

        isAscomAutoConnect = self.ui.checkASCOMAutoConnect.isChecked()

        for driver in self.drivers:

            if driver not in self.driversData:
                continue

            isValid = self.driversData[driver]['framework'] != ''

            if not isValid:
                continue

            isAscom = self.driversData[driver]['framework'] == 'ascom'

            if isAscom and isAscomAutoConnect:
                autoStart = True

            elif isAscom and not isAscomAutoConnect:
                autoStart = False

            else:
                autoStart = True

            self.startDriver(driver=driver, autoStart=autoStart)

        return True

    def manualStopAllAscomDrivers(self):
        """

        :return: True for test purpose
        """
        for driver in self.drivers:

            if driver not in self.driversData:
                continue

            isAscom = self.driversData[driver]['framework'] == 'ascom'

            if isAscom:
                self.stopDriver(driver=driver)

        return True

    def manualStartAllAscomDrivers(self):
        """

        :return: True for test purpose
        """

        for driver in self.drivers:

            if driver not in self.driversData:
                continue

            isAscom = self.driversData[driver]['framework'] == 'ascom'

            if isAscom:
                self.startDriver(driver=driver, autoStart=True)

        return True

    def dispatchDriverDropdown(self):
        """
        dispatchDriverDropdown maps the gui event received from signals to the methods doing
        the real stuff. this splits function and gui reaction into two separate tasks.
        if a dropDown action is taken, this means and new driver has been selected,
        so the old one has to be stopped, the new configured and started.

        :return: true for test purpose
        """

        sender = self.sender()
        isDisabled = sender.currentText() == 'device disabled'
        driver = self.returnDriver(sender, self.drivers, addKey='uiDropDown')

        if driver:
            if isDisabled:
                framework = ''
            else:
                framework = sender.currentText().split('-')[0].rstrip()

            self.driversData[driver]['framework'] = framework
            self.stopDriver(driver=driver)

            if framework:
                self.startDriver(driver=driver)

        return True

    def scanValid(self, driver=None, deviceName=''):
        """
        scanValid checks if the calling device fits to the summary of all devices and gives
        back if it should be skipped

        :param deviceName:
        :param driver:
        :return:
        """

        if not driver:
            return False
        if not deviceName:
            return False

        if hasattr(self.drivers[driver]['class'], 'signals'):
            if self.sender() != self.drivers[driver]['class'].signals:
                return False

        else:
            driverClass = self.drivers[driver]['class']

            if not driverClass.framework:
                return False

            if driverClass.run[driverClass.framework].deviceName != deviceName:
                return False

        return True

    def serverDisconnected(self, deviceList):
        """
        serverDisconnected writes info to message window and recolors the status

        :param deviceList:
        :return: true for test purpose
        """

        if not deviceList:
            return False

        deviceName = list(deviceList.keys())[0]

        for driver in self.drivers:
            if not self.scanValid(driver=driver, deviceName=deviceName):
                continue

            self.drivers[driver]['uiDropDown'].setStyleSheet(self.BACK_NORM)
            # self.app.message.emit(f'Disconnected server: [{driver}] ', 0)
        return True

    def deviceConnected(self, deviceName):
        """
        showCoverDeviceConnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :param deviceName:
        :return: true for test purpose
        """
        if not deviceName:
            return False

        for driver in self.drivers:
            if not self.scanValid(driver=driver, deviceName=deviceName):
                continue

            self.drivers[driver]['uiDropDown'].setStyleSheet(self.BACK_GREEN)
            self.deviceStat[driver] = True
            # self.app.message.emit(f'{driver} connected', 0)
        return True

    def deviceDisconnected(self, deviceName):
        """
        showCoverDeviceDisconnected changes the style of related ui groups to make it clear
        to the user, which function is actually available

        :param deviceName:
        :return: true for test purpose
        """

        for driver in self.drivers:
            if not self.scanValid(driver=driver, deviceName=deviceName):
                continue

            self.drivers[driver]['uiDropDown'].setStyleSheet(self.BACK_NORM)
            self.deviceStat[driver] = False
            # self.app.message.emit(f'{driver} disconnected', 0)
        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries
import json
import os
from threading import Lock

# external packages
from matplotlib import pyplot as plt
from matplotlib import ticker
from matplotlib.colors import Normalize
import numpy as np
from scipy.stats.mstats import winsorize

# local import
from gui.utilities import toolsQtWidget
from gui.widgets import analyse_ui
from base.tpool import Worker


class AnalyseWindow(toolsQtWidget.MWidget):
    """
    the analyse window class handles

    """

    __all__ = ['AnalyseWindow',
               ]

    def __init__(self, app):
        super().__init__()
        self.app = app

        self.ui = analyse_ui.Ui_AnalyseDialog()
        self.ui.setupUi(self)
        self.initUI()
        self.threadPool = app.threadPool
        self.threadCounter = 0
        self.threadingLock = Lock()
        self.workers = []

        self.latitude = None
        self.pierside = None
        self.countSequence = None
        self.index = None
        self.scaleS = None
        self.altitude = None
        self.azimuth = None
        self.errorAngle = None
        self.errorRMS = None
        self.errorRA_S = None
        self.errorDEC_S = None
        self.errorRA = None
        self.errorDEC = None
        self.angularPosRA = None
        self.angularPosDEC = None

        self.wIcon(self.ui.load, 'load')

        self.raPointErrorsRaw = self.embedMatplot(self.ui.raPointErrorsRaw)
        self.decPointErrorsRaw = self.embedMatplot(self.ui.decPointErrorsRaw)
        self.raPointErrorsRawRef = self.embedMatplot(self.ui.raPointErrorsRawRef)
        self.decPointErrorsRawRef = self.embedMatplot(self.ui.decPointErrorsRawRef)
        self.raModelErrors = self.embedMatplot(self.ui.raModelErrors)
        self.decModelErrors = self.embedMatplot(self.ui.decModelErrors)
        self.raModelErrorsRef = self.embedMatplot(self.ui.raModelErrorsRef)
        self.decModelErrorsRef = self.embedMatplot(self.ui.decModelErrorsRef)
        self.scaleImage = self.embedMatplot(self.ui.scaleImage)
        self.modelPositions = self.embedMatplot(self.ui.modelPositions)
        self.errorAscending = self.embedMatplot(self.ui.errorAscending)
        self.errorDistribution = self.embedMatplot(self.ui.errorDistribution)

        self.ui.load.clicked.connect(self.loadModel)
        self.ui.winsorizedLimit.clicked.connect(self.drawAll)

        self.charts = [self.draw_raPointErrorsRaw,
                       self.draw_decPointErrorsRaw,
                       self.draw_raModelErrors,
                       self.draw_decModelErrors,
                       self.draw_raModelErrorsRef,
                       self.draw_decModelErrorsRef,
                       self.draw_raPointErrorsRawRef,
                       self.draw_decPointErrorsRawRef,
                       self.draw_scaleImage,
                       self.draw_modelPositions,
                       self.draw_errorAscending,
                       self.draw_errorDistribution]

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """

        if 'analyseW' not in self.app.config:
            self.app.config['analyseW'] = {}
        config = self.app.config['analyseW']
        x = config.get('winPosX', 40)
        y = config.get('winPosY', 40)

        if x > self.screenSizeX:
            x = 0

        if y > self.screenSizeY:
            y = 0

        self.move(x, y)
        height = config.get('height', 600)
        width = config.get('width', 800)
        self.resize(width, height)
        self.ui.winsorizedLimit.setChecked(config.get('winsorizedLimit', False))

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """
        if 'analyseW' not in self.app.config:
            self.app.config['analyseW'] = {}

        config = self.app.config['analyseW']
        config['winPosX'] = self.pos().x()
        config['winPosY'] = self.pos().y()
        config['height'] = self.height()
        config['width'] = self.width()
        config['winsorizedLimit'] = self.ui.winsorizedLimit.isChecked()

        return True

    def closeEvent(self, closeEvent):
        """
        closeEvent is overloaded to be able to store the data before the windows is close
        and all it's data is garbage collected

        :param closeEvent:
        :return:
        """
        self.storeConfig()
        self.app.showAnalyse.disconnect(self.showAnalyse)

        # gui signals
        super().closeEvent(closeEvent)

    def resizeEvent(self, event):
        """
        we are using the resize event to reset the timer, which means waiting for
        RESIZE_FINISHED_TIMEOUT in total before redrawing the complete hemisphere.
        as we are using a 0.1s cyclic timer.

        :param event:
        :return:
        """

        super().resizeEvent(event)

    def showWindow(self):
        """
        showWindow starts constructing the main window for satellite view and shows the
        window content

        :return: True for test purpose
        """

        self.show()
        self.app.showAnalyse.connect(self.showAnalyse)

        return True

    def writeGui(self, data, loadFilePath):
        """

        :param data:
        :param loadFilePath:
        :return: True for test purpose
        """

        d = data[0]
        de = data[-1]

        self.ui.filename.setText(os.path.basename(loadFilePath))
        self.ui.eposure.setText(f'{d.get("exposureTime", "")}')
        self.ui.solver.setText(d.get('astrometryApp', ""))
        self.ui.binning.setText(f'{d.get("binning", 0):1.0f}')
        self.ui.time.setText(d.get('julianDate', '').replace('T', '  ').replace('Z', ''))
        self.ui.subframe.setText(f'{d.get("subFrame", 0):3.0f}')
        self.ui.modelTerms.setText(f'{d.get("modelTerms", 0):02.0f}')
        self.ui.modelErrorRMS.setText(f'{d.get("modelErrorRMS", 0):5.1f}')
        self.ui.modelOrthoError.setText(f'{d.get("modelOrthoError", 0):5.0f}')
        self.ui.modelPolarError.setText(f'{d.get("modelPolarError", 0):5.0f}')
        self.ui.focalLength.setText(f'{d.get("focalLength", 0):4.0f}')
        self.ui.profile.setText(f'{d.get("profile", "")}')
        self.ui.firmware.setText(f'{d.get("firmware", "")}')
        self.ui.totalPoints.setText(f'{d.get("lenSequence", 0)}')
        self.ui.goodPoints.setText(f'{de.get("errorIndex", 0)}')
        text = f'{d.get("mirroredS", "")}' + f'{d.get("flippedS", "")}'
        self.ui.mirrored.setText(text)
        version = d.get("version", "").lstrip('MountWizzard4 - v')
        self.ui.version.setText(f'{version}')

        return True

    def generateDataSets(self, modelJSON):
        """

        :param modelJSON:
        :return:
        """
        model = dict()

        for key in modelJSON[0].keys():
            model[key] = list()
            for index in range(0, len(modelJSON)):
                model[key].append(modelJSON[index][key])

        self.latitude = modelJSON[0].get('latitude')
        self.pierside = np.asarray(model['pierside'])
        self.countSequence = np.asarray(model['countSequence'])
        self.index = np.asarray(model['errorIndex']) - 1
        self.scaleS = np.asarray(model['scaleS'])
        self.altitude = np.asarray(model['altitude'])
        self.azimuth = np.asarray(model['azimuth'])
        self.errorAngle = np.asarray(model['errorAngle'])
        self.errorRMS = np.asarray(model['errorRMS'])
        self.errorRA_S = np.asarray(model['errorRA_S'])
        self.errorDEC_S = np.asarray(model['errorDEC_S'])
        self.errorRA = np.asarray(model['errorRA'])
        self.errorDEC = np.asarray(model['errorDEC'])
        self.angularPosRA = np.asarray(model['angularPosRA'])
        self.angularPosDEC = np.asarray(model['angularPosDEC'])

        return True

    def processModel(self, loadFilePath):
        """
        processModel selects one or more models from the files system

        :return: success
        """

        with open(loadFilePath, 'r') as infile:
            modelJSON = json.load(infile)

        self.writeGui(modelJSON, loadFilePath)
        self.generateDataSets(modelJSON)
        self.drawAll()

        return True

    def loadModel(self):
        """
        loadModel selects one or more models from the files system

        :return: success
        """

        folder = self.app.mwGlob['modelDir']
        val = self.openFile(self, 'Open model file', folder, 'Model files (*.model)',
                            multiple=False,
                            reverseOrder=True)
        loadFilePath, fileName, ext = val

        if loadFilePath:
            self.processModel(loadFilePath)

        return True

    def showAnalyse(self, path):
        """

        :param path:
        :return: True for test purpose
        """

        if path:
            self.processModel(path)

        return True

    def plotFigureFlat(self, axe, x, y, p, xLabel, yLabel, sort=False, poly=0):
        """

        :param axe:
        :param x:
        :param y:
        :param p: pierside
        :param xLabel:
        :param yLabel:
        :param sort:
        :param poly:
        :return: True for test purpose
        """

        axe.set_xlabel(xLabel)
        axe.set_ylabel(yLabel)

        if sort:
            x, y, pierside = zip(*sorted(zip(x, y, p)))

        else:
            pierside = p

        if self.ui.winsorizedLimit.isChecked():
            y = winsorize(y, limits=0.03)

        xW = [x for x, y, p in zip(x, y, pierside) if p == 'W']
        yW = [y for x, y, p in zip(x, y, pierside) if p == 'W']
        xE = [x for x, y, p in zip(x, y, pierside) if p == 'E']
        yE = [y for x, y, p in zip(x, y, pierside) if p == 'E']

        if poly:
            try:
                polynomW = np.polyfit(xW, yW, poly)

            except Exception as e:
                self.log.debug(f'Interpolation failed: {e}')

            else:
                hasNoNan = not np.isnan(polynomW).any()
                hasNoInf = not np.isinf(polynomW).any()

                if hasNoNan and hasNoInf:
                    meanW = np.poly1d(polynomW)(xW)
                    axe.plot(xW, meanW, color=self.M_GREEN, alpha=0.4, lw=4)

            try:
                polynomE = np.polyfit(xE, yE, poly)

            except Exception as e:
                self.log.debug(f'Interpolation failed: {e}')

            else:
                hasNoNan = not np.isnan(polynomE).any()
                hasNoInf = not np.isinf(polynomE).any()

                if hasNoNan and hasNoInf:
                    meanE = np.poly1d(polynomE)(xE)
                    axe.plot(xE, meanE, color=self.M_YELLOW, alpha=0.4, lw=4)

        axe.plot(xW, yW, marker='.', markersize=7, linestyle='none', color=self.M_GREEN)
        axe.plot(xE, yE, marker='.', markersize=7, linestyle='none', color=self.M_YELLOW)

        axe.figure.canvas.draw()

        return True

    def draw_raPointErrorsRaw(self):
        """
        draw_raPointErrorsRaw draws a plot of

        :return:    True if ok for testing
        """

        axe, _ = self.generateFlat(widget=self.raPointErrorsRaw)

        x = self.index
        y = self.errorRA_S
        p = self.pierside
        xLabel = 'Star Number'
        yLabel = 'Error per Star [arcsec]'

        self.plotFigureFlat(axe, x, y, p, xLabel, yLabel, False, 3)

        return True

    def draw_decPointErrorsRaw(self):
        """
        draw_decPointErrorsRaw draws a plot of raw errors in dec. Please watch the inverse sign
        for pierside east.

        :return:    True if ok for testing
        """

        axe, _ = self.generateFlat(widget=self.decPointErrorsRaw)

        x = self.index
        y = self.errorDEC_S
        p = self.pierside
        y = [y if p == 'W' else -y for y, p in zip(y, p)]
        xLabel = 'Star Number'
        yLabel = 'Error per Star [arcsec]'

        self.plotFigureFlat(axe, x, y, p, xLabel, yLabel, False, 3)

        return True

    def draw_raModelErrors(self):
        """
        draw_raModelErrors draws a plot of

        :return:    True if ok for testing
        """

        axe, _ = self.generateFlat(widget=self.raModelErrors)

        x = self.index
        y = self.errorRA
        p = self.pierside
        xLabel = 'Star Number'
        yLabel = 'Error per Star [arcsec]'

        self.plotFigureFlat(axe, x, y, p, xLabel, yLabel, False, 3)

        return True

    def draw_decModelErrors(self):
        """
        draw_decModelErrors draws a plot of

        :return:    True if ok for testing
        """

        axe, _ = self.generateFlat(widget=self.decModelErrors)

        x = self.index
        y = self.errorDEC
        p = self.pierside
        xLabel = 'Star Number'
        yLabel = 'Error per Star [arcsec]'

        self.plotFigureFlat(axe, x, y, p, xLabel, yLabel, False, 3)

        return True

    def draw_raModelErrorsRef(self):
        """
        draw_raModelErrorsRef draws a plot of

        :return:    True if ok for testing
        """

        axe, _ = self.generateFlat(widget=self.raModelErrorsRef)

        x = self.angularPosRA
        y = self.errorRA
        p = self.pierside
        xLabel = 'RA Encoder Abs [deg]'
        yLabel = 'Error per Star [arcsec]'

        self.plotFigureFlat(axe, x, y, p, xLabel, yLabel, True, 3)

        return True

    def draw_decModelErrorsRef(self):
        """
        draw_decModelErrorsRef draws a plot of

        :return:    True if ok for testing
        """

        axe, _ = self.generateFlat(widget=self.decModelErrorsRef)

        x = self.angularPosDEC
        y = self.errorDEC
        p = self.pierside
        y = [x if p == 'W' else -x for x, p in zip(y, p)]
        xLabel = 'DEC Encoder Abs [deg]'
        yLabel = 'Error per Star [arcsec]'

        self.plotFigureFlat(axe, x, y, p, xLabel, yLabel, True, 3)

        return True

    def draw_raPointErrorsRawRef(self):
        """
        draw_raPointErrorsRawRef draws a plot of

        :return:    True if ok for testing
        """

        axe, _ = self.generateFlat(widget=self.raPointErrorsRawRef)

        x = self.angularPosRA
        y = self.errorRA_S
        p = self.pierside
        xLabel = 'RA Encoder Abs [deg]'
        yLabel = 'Error per Star [arcsec]'

        self.plotFigureFlat(axe, x, y, p, xLabel, yLabel, True, 3)

        return True

    def draw_decPointErrorsRawRef(self):
        """
        draw_decPointErrorsRawRef draws a plot of

        :return:    True if ok for testing
        """

        axe, _ = self.generateFlat(widget=self.decPointErrorsRawRef)

        x = self.angularPosDEC
        y = self.errorDEC_S
        p = self.pierside
        y = [y if p == 'W' else -y for y, p in zip(y, p)]
        xLabel = 'DEC Encoder Abs [deg]'
        yLabel = 'Error per Star [arcsec]'

        self.plotFigureFlat(axe, x, y, p, xLabel, yLabel, True, 3)

        return True

    def draw_scaleImage(self):
        """
        draw_raPointErrorsRaw draws a plot of

        :return:    True if ok for testing
        """

        axe, _ = self.generateFlat(widget=self.scaleImage)
        axe.get_yaxis().set_major_formatter(ticker.FormatStrFormatter('%.3f',))

        x = self.index
        y = self.scaleS
        p = self.pierside
        xLabel = 'Star Number'
        yLabel = 'Image Scale [arcsec/pix]'

        self.plotFigureFlat(axe, x, y, p, xLabel, yLabel, False, 3)

        return True

    def draw_errorAscending(self):
        """
        showErrorAscending draws a plot of the align model stars and their errors in ascending
        order.

        :return:    True if ok for testing
        """

        axe, fig = self.generateFlat(widget=self.errorAscending)

        xLabel = 'Star Number'
        yLabel = 'Error per Star [arcsec]'

        y = self.errorRMS
        pierside = self.pierside
        x = self.index

        temp = sorted(zip(y, pierside))

        y = [x[0] for x in temp]
        p = [x[1] for x in temp]

        self.plotFigureFlat(axe, x, y, p, xLabel, yLabel, False, 3)

        return True

    def draw_modelPositions(self):
        """
        showModelPosition draws a polar plot of the align model stars and their errors in
        color. the basic setup of the plot is taking place in the central widget class.
        which is instantiated from there. important: the coordinate in model is in HA and
        DEC  and not in RA and DEC. using skyfield is a little bit misleading, because you
        address the hour angle as .ra.hours

        the vectors displayed are derived from the spec of 10micron:
            ppp is the polar angle of the measured star with respect to the modeled star
            in the equatorial system in degrees from 0 to 359 (0 towards the north pole,
            90 towards east)

        :return:    True if ok for testing
        """

        axe, fig = self.generatePolar(widget=self.modelPositions)

        axe.set_ylim(0, 90)
        axe.set_yticklabels('')

        altitude = self.altitude
        azimuth = self.azimuth
        error = self.errorRMS
        ang = self.errorAngle

        # and plot it
        cm = plt.cm.get_cmap('RdYlGn_r')
        sMax = max(error)
        sMin = min(error)
        az = azimuth / 180.0 * np.pi
        alt = 90 - altitude

        scatter = axe.scatter(az, alt, c=error, vmin=sMin, vmax=sMax, cmap=cm)

        norm = Normalize(vmin=sMin, vmax=sMax)
        sm = plt.cm.ScalarMappable(cmap=cm, norm=norm)
        sm.set_array([])

        fm = ticker.FormatStrFormatter('%1.0f')
        cb = plt.colorbar(sm, pad=0.1, fraction=0.12, aspect=25, shrink=0.9, format=fm)
        cb.set_label('Error [arcsec]', color=self.M_BLUE)
        yTicks = plt.getp(cb.ax.axes, 'yticklabels')
        plt.setp(yTicks, color=self.M_BLUE, fontweight='bold')

        lat = self.latitude

        if lat is None:
            axe.figure.canvas.draw()
            return True

        npX = (90 - lat) * np.cos(np.radians(0 + 90))
        npY = (90 - lat) * np.sin(np.radians(0 + 90))

        axe.plot(0, 90 - lat, marker='o', markersize=10, color=self.M_BLUE, alpha=0.8,
                 zorder=-10)
        axe.plot(0, 90 - lat, marker='o', markersize=20, color=self.M_BLUE, alpha=0.8,
                 lw=10, fillstyle='none', zorder=-10)
        axe.plot(0, 90 - lat, marker='o', markersize=35, color=self.M_BLUE, alpha=0.8,
                 lw=10, fillstyle='none', zorder=-10)

        for alt, az, ang, err in zip(altitude, azimuth, ang, error):
            pX = (90 - alt) * np.cos(np.radians(az + 90))
            pY = (90 - alt) * np.sin(np.radians(az + 90))

            vec = np.arctan2(pY - npY, pX - npX) + np.radians(90) + np.radians(ang)

            x = az / 180.0 * np.pi
            y = 90 - alt
            u = np.sin(vec)
            v = np.cos(vec)

            col = cm((err - sMin) / sMax)

            axe.quiver(x, y, u, v,
                       color=col,
                       scale=17,
                       headlength=0,
                       headwidth=1,
                       alpha=0.8,
                       zorder=-10,
                       )

        formatString = ticker.FormatStrFormatter('%1.0f')
        colorbar = fig.colorbar(scatter,
                                pad=0.1,
                                fraction=0.12,
                                aspect=25,
                                shrink=0.9,
                                format=formatString,
                                )

        colorbar.set_label('Error [arcsec]', color=self.M_BLUE)
        yTicks = plt.getp(colorbar.ax.axes, 'yticklabels')
        plt.setp(yTicks, color=self.M_BLUE, fontweight='bold')

        axe.figure.canvas.draw()

        return True

    def draw_errorDistribution(self):
        """
        showErrorDistribution draws a polar plot of the align model stars and their errors in
        color. the basic setup of the plot is taking place in the central widget class.
        which is instantiated from there. important: the coordinate in model is in HA and
        DEC  and not in RA and DEC. using skyfield is a little bit misleading, because you
        address the hour angle as .ra.hours

        :return:    True if ok for testing
        """

        axe, fig = self.generatePolar(widget=self.errorDistribution)

        x = [val / 180.0 * np.pi for val in self.errorAngle]
        y = self.errorRMS

        for x, y, pierside in zip(x, y, self.pierside):
            if pierside == 'W':
                color = self.M_GREEN

            else:
                color = self.M_YELLOW

            axe.plot(x, y, marker='.', markersize=7, linestyle='none', color=color)

        axe.figure.canvas.draw()

        return True

    def workerDrawAll(self):
        """

        :return:
        """
        for chart in self.charts:
            chart()

        return True

    def drawAll(self):
        """

        :return: true for test purpose
        """

        if self.countSequence is None:
            return False

        worker = Worker(self.workerDrawAll)
        self.threadPool.start(worker)

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages

# local imports
from base.ascomClass import AscomClass


class SensorWeatherAscom(AscomClass):

    __all__ = ['SensorWeatherAscom',
               ]

    CYCLE_DEVICE = 3000
    CYCLE_DATA = 1000

    def __init__(self, app=None, signals=None, data=None):
        super().__init__(app=app, data=data, threadPool=app.threadPool)

        self.signals = signals
        self.data = data

    def getInitialConfig(self):
        """

        :return: true for test purpose
        """

        super().getInitialConfig()

        return True

    def workerPollData(self):
        """

        :return: true for test purpose
        """

        if not self.deviceConnected:
            return False

        self.dataEntry(self.client.temperature, 'WEATHER_PARAMETERS.WEATHER_TEMPERATURE')
        self.dataEntry(self.client.pressure, 'WEATHER_PARAMETERS.WEATHER_PRESSURE')
        self.dataEntry(self.client.dewpoint, 'WEATHER_PARAMETERS.WEATHER_DEWPOINT')
        self.dataEntry(self.client.humidity, 'WEATHER_PARAMETERS.WEATHER_HUMIDITY')

        try:
            val = self.client.skyquality
        except Exception:
            val = 0
        self.dataEntry(val, 'SKY_QUALITY.SKY_BRIGHTNESS')

        try:
            val = self.client.cloudcover
        except Exception:
            val = 0
        self.dataEntry(val, 'cloudCover')

        try:
            val = self.client.rainrate
        except Exception:
            val = 0
        self.dataEntry(val, 'rain')

        try:
            val = self.client.winddirection
        except Exception:
            val = 0
        self.dataEntry(val, 'windDir')

        try:
            val = self.client.windspeed
        except Exception:
            val = 0
        self.dataEntry(val, 'windSpeed')

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages

# local import


class BuildPoints:
    """
    """

    def __init__(self):
        self.lastGenerator = 'none'

        self.ui.genBuildGrid.clicked.connect(self.genBuildGrid)
        self.ui.genBuildAlign3.clicked.connect(self.genBuildAlign3)
        self.ui.genBuildAlign6.clicked.connect(self.genBuildAlign6)
        self.ui.genBuildAlign9.clicked.connect(self.genBuildAlign9)
        self.ui.numberGridPointsCol.valueChanged.connect(self.genBuildGrid)
        self.ui.numberGridPointsRow.valueChanged.connect(self.genBuildGrid)
        self.ui.altitudeMin.valueChanged.connect(self.genBuildGrid)
        self.ui.altitudeMax.valueChanged.connect(self.genBuildGrid)
        self.ui.genBuildMax.clicked.connect(self.genBuildMax)
        self.ui.genBuildMed.clicked.connect(self.genBuildMed)
        self.ui.genBuildNorm.clicked.connect(self.genBuildNorm)
        self.ui.genBuildMin.clicked.connect(self.genBuildMin)
        self.ui.genBuildFile.clicked.connect(self.genBuildFile)
        self.ui.genBuildDSO.clicked.connect(self.genBuildDSO)
        self.ui.numberDSOPoints.valueChanged.connect(self.genBuildDSO)
        self.ui.durationDSO.valueChanged.connect(self.genBuildDSO)
        self.ui.timeShiftDSO.valueChanged.connect(self.genBuildDSO)
        self.ui.saveBuildPoints.clicked.connect(self.saveBuildFile)
        self.ui.saveBuildPointsAs.clicked.connect(self.saveBuildFileAs)
        self.ui.loadBuildPoints.clicked.connect(self.loadBuildFile)
        self.ui.genBuildSpiralMax.clicked.connect(self.genBuildSpiralMax)
        self.ui.genBuildSpiralMed.clicked.connect(self.genBuildSpiralMed)
        self.ui.genBuildSpiralNorm.clicked.connect(self.genBuildSpiralNorm)
        self.ui.genBuildSpiralMin.clicked.connect(self.genBuildSpiralMin)
        self.ui.clearBuildP.clicked.connect(self.clearBuildP)
        self.ui.checkSortNothing.clicked.connect(self.processPoints)
        self.ui.checkSortEW.clicked.connect(self.processPoints)
        self.ui.checkSortHL.clicked.connect(self.processPoints)
        self.ui.checkAvoidFlip.clicked.connect(self.processPoints)
        self.ui.checkAutoDeleteMeridian.clicked.connect(self.autoDeletePoints)
        self.ui.checkAutoDeleteHorizon.clicked.connect(self.autoDeletePoints)

    def initConfig(self):
        """
        initConfig read the key out of the configuration dict and stores it to the gui
        elements. if some initialisations have to be proceeded with the loaded persistent
        data, they will be launched as well in this method.

        :return: True for test purpose
        """

        config = self.app.config['mainW']

        self.ui.numberGridPointsCol.valueChanged.disconnect(self.genBuildGrid)
        self.ui.numberGridPointsRow.valueChanged.disconnect(self.genBuildGrid)
        self.ui.altitudeMin.valueChanged.disconnect(self.genBuildGrid)
        self.ui.altitudeMax.valueChanged.disconnect(self.genBuildGrid)
        self.ui.numberDSOPoints.valueChanged.disconnect(self.genBuildDSO)
        self.ui.durationDSO.valueChanged.disconnect(self.genBuildDSO)
        self.ui.timeShiftDSO.valueChanged.disconnect(self.genBuildDSO)

        self.ui.buildPFileName.setText(config.get('buildPFileName', ''))
        self.ui.numberGridPointsRow.setValue(config.get('numberGridPointsRow', 5))
        self.ui.numberGridPointsCol.setValue(config.get('numberGridPointsCol', 6))
        self.ui.altitudeMin.setValue(config.get('altitudeMin', 30))
        self.ui.altitudeMax.setValue(config.get('altitudeMax', 75))
        self.ui.numberDSOPoints.setValue(config.get('numberDSOPoints', 15))
        self.ui.durationDSO.setValue(config.get('durationDSO', 7))
        self.ui.timeShiftDSO.setValue(config.get('timeShiftDSO', 0))

        self.ui.checkAutoDeleteMeridian.setChecked(config.get('checkAutoDeleteMeridian', False))
        self.ui.checkAutoDeleteHorizon.setChecked(config.get('checkAutoDeleteHorizon', True))
        self.ui.checkSafetyMarginHorizon.setChecked(config.get('checkSafetyMarginHorizon',
                                                               False))
        self.ui.safetyMarginHorizon.setValue(config.get('safetyMarginHorizon', 0))
        self.ui.checkAvoidFlip.setChecked(config.get('checkAvoidFlip', False))
        self.ui.checkSortNothing.setChecked(config.get('checkSortNothing', True))
        self.ui.checkSortEW.setChecked(config.get('checkSortEW', False))
        self.ui.checkSortHL.setChecked(config.get('checkSortHL', False))
        self.ui.keepGeneratedPoints.setChecked(config.get('keepGeneratedPoints', False))

        self.ui.numberGridPointsCol.valueChanged.connect(self.genBuildGrid)
        self.ui.numberGridPointsRow.valueChanged.connect(self.genBuildGrid)
        self.ui.altitudeMin.valueChanged.connect(self.genBuildGrid)
        self.ui.altitudeMax.valueChanged.connect(self.genBuildGrid)
        self.ui.numberDSOPoints.valueChanged.connect(self.genBuildDSO)
        self.ui.durationDSO.valueChanged.connect(self.genBuildDSO)
        self.ui.timeShiftDSO.valueChanged.connect(self.genBuildDSO)

        return True

    def storeConfig(self):
        """
        storeConfig writes the keys to the configuration dict and stores. if some
        saving has to be proceeded to persistent data, they will be launched as
        well in this method.

        :return: True for test purpose
        """

        config = self.app.config['mainW']
        config['buildPFileName'] = self.ui.buildPFileName.text()
        config['numberGridPointsRow'] = self.ui.numberGridPointsRow.value()
        config['numberGridPointsCol'] = self.ui.numberGridPointsCol.value()
        config['altitudeMin'] = self.ui.altitudeMin.value()
        config['altitudeMax'] = self.ui.altitudeMax.value()
        config['numberDSOPoints'] = self.ui.numberDSOPoints.value()
        config['durationDSO'] = self.ui.durationDSO.value()
        config['timeShiftDSO'] = self.ui.timeShiftDSO.value()
        config['checkAutoDeleteMeridian'] = self.ui.checkAutoDeleteMeridian.isChecked()
        config['checkAutoDeleteHorizon'] = self.ui.checkAutoDeleteHorizon.isChecked()
        config['checkSafetyMarginHorizon'] = self.ui.checkSafetyMarginHorizon.isChecked()
        config['safetyMarginHorizon'] = self.ui.safetyMarginHorizon.value()
        config['checkAvoidFlip'] = self.ui.checkAvoidFlip.isChecked()
        config['checkSortNothing'] = self.ui.checkSortNothing.isChecked()
        config['checkSortEW'] = self.ui.checkSortEW.isChecked()
        config['checkSortHL'] = self.ui.checkSortHL.isChecked()
        config['keepGeneratedPoints'] = self.ui.keepGeneratedPoints.isChecked()

        return True

    def genBuildGrid(self):
        """
        genBuildGrid generates a grid of point for model build based on gui data. the cols
        have to be on even numbers.

        :return: success
        """

        self.lastGenerator = 'grid'

        self.ui.numberGridPointsRow.setEnabled(False)
        self.ui.numberGridPointsCol.setEnabled(False)
        self.ui.altitudeMin.setEnabled(False)
        self.ui.altitudeMax.setEnabled(False)

        row = self.ui.numberGridPointsRow.value()
        col = self.ui.numberGridPointsCol.value()

        # we only have equal cols
        col = 2 * int(col / 2)
        self.ui.numberGridPointsCol.setValue(col)
        minAlt = self.ui.altitudeMin.value()
        maxAlt = self.ui.altitudeMax.value()
        keep = self.ui.keepGeneratedPoints.isChecked()

        suc = self.app.data.genGrid(minAlt=minAlt,
                                    maxAlt=maxAlt,
                                    numbRows=row,
                                    numbCols=col,
                                    keep=keep)

        if not suc:
            self.ui.numberGridPointsRow.setEnabled(True)
            self.ui.numberGridPointsCol.setEnabled(True)
            self.ui.altitudeMin.setEnabled(True)
            self.ui.altitudeMax.setEnabled(True)
            self.app.message.emit('Could not generate grid', 2)
            return False

        self.processPoints()

        self.ui.numberGridPointsRow.setEnabled(True)
        self.ui.numberGridPointsCol.setEnabled(True)
        self.ui.altitudeMin.setEnabled(True)
        self.ui.altitudeMax.setEnabled(True)

        return True

    def genBuildAlign3(self):
        """
        genBuildAlign3 generates a grid of 3 point for model build based on gui data.

        :return: success
        """

        self.lastGenerator = 'align3'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.genAlign(altBase=55,
                                     azBase=10,
                                     numberBase=3,
                                     keep=keep)

        if not suc:
            self.app.message.emit('Could not generate 3 align stars', 2)
            return False

        self.processPoints()

        return True

    def genBuildAlign6(self):
        """
        genBuildAlign6 generates a grid of 6 point for model build based on gui data.

        :return: success
        """

        self.lastGenerator = 'align6'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.genAlign(altBase=55,
                                     azBase=10,
                                     numberBase=6,
                                     keep=keep)

        if not suc:
            self.app.message.emit('Could not generate 6 align stars', 2)
            return False

        self.processPoints()

        return True

    def genBuildAlign9(self):
        """
        genBuildAlign9 generates a grid of 9 point for model build based on gui data.

        :return: success
        """

        self.lastGenerator = 'align9'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.genAlign(altBase=55,
                                     azBase=10,
                                     numberBase=9,
                                     keep=keep)

        if not suc:
            self.app.message.emit('Could not generate 9 align stars', 2)
            return False

        self.processPoints()

        return True

    def genBuildMax(self):
        """
        genBuildMax generates the point pattern based on greater circles for model build.
        the point are calculated for the observers position. max goes for approx 100 points
        effectively when removing the horizon.

        :return: success
        """

        self.lastGenerator = 'max'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.genGreaterCircle(selection='max', keep=keep)

        if not suc:
            self.app.message.emit('Build points [max] cannot be generated', 2)
            return False

        self.processPoints()

        return True

    def genBuildMed(self):
        """
        genBuildMed generates the point pattern based on greater circles for model build.
        the point are calculated for the observers position. max goes for approx 70 points
        effectively when removing the horizon.

        :return: success
        """

        self.lastGenerator = 'med'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.genGreaterCircle(selection='med', keep=keep)

        if not suc:
            self.app.message.emit('Build points [med] cannot be generated', 2)
            return False

        self.processPoints()

        return True

    def genBuildNorm(self):
        """
        genBuildNorm generates the point pattern based on greater circles for model build.
        the point are calculated for the observers position. max goes for approx 40 points
        effectively when removing the horizon.

        :return: success
        """

        self.lastGenerator = 'norm'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.genGreaterCircle(selection='norm', keep=keep)

        if not suc:
            self.app.message.emit('Build points [norm] cannot be generated', 2)
            return False

        self.processPoints()

        return True

    def genBuildMin(self):
        """
        genBuildMin generates the point pattern based on greater circles for model build.
        the point are calculated for the observers position. min goes for approx 25 points
        effectively when removing the horizon.

        :return: success
        """

        self.lastGenerator = 'min'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.genGreaterCircle(selection='min', keep=keep)

        if not suc:
            self.app.message.emit('Build points [min] cannot be generated', 2)
            return False

        self.processPoints()

        return True

    def genBuildDSO(self):
        """
        genBuildDSO generates points along the actual tracking path

        :return: success
        """

        self.lastGenerator = 'dso'
        ra = self.app.mount.obsSite.raJNow
        dec = self.app.mount.obsSite.decJNow
        timeJD = self.app.mount.obsSite.timeJD
        location = self.app.mount.obsSite.location

        if ra is None or dec is None or location is None:
            self.app.message.emit('DSO Path cannot be generated', 2)
            return False

        self.ui.numberDSOPoints.setEnabled(False)
        self.ui.durationDSO.setEnabled(False)
        self.ui.timeShiftDSO.setEnabled(False)

        numberPoints = self.ui.numberDSOPoints.value()
        duration = self.ui.durationDSO.value()
        timeShift = self.ui.timeShiftDSO.value()

        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.generateDSOPath(ra=ra,
                                            dec=dec,
                                            timeJD=timeJD,
                                            location=location,
                                            numberPoints=numberPoints,
                                            duration=duration,
                                            timeShift=timeShift,
                                            keep=keep,
                                            )

        if not suc:
            self.ui.numberDSOPoints.setEnabled(True)
            self.ui.durationDSO.setEnabled(True)
            self.ui.timeShiftDSO.setEnabled(True)
            self.app.message.emit('DSO Path cannot be generated', 2)
            return False

        self.processPoints()

        self.ui.numberDSOPoints.setEnabled(True)
        self.ui.durationDSO.setEnabled(True)
        self.ui.timeShiftDSO.setEnabled(True)

        return True

    def genBuildSpiralMax(self):
        """
        genBuildGoldenSpiral generates points along the actual tracking path
        as the processing might take to long (at least on ubuntu), we have to find a
        workaround to this behavior:
        https://stackoverflow.com/questions/41568990/
        how-do-i-prevent-double-valuechanged-events-when-i-press-the-arrows-in-a-qspinbo

        :return: success
        """

        self.lastGenerator = 'spiralMax'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.generateGoldenSpiral(numberPoints=350, keep=keep)

        if not suc:
            self.app.message.emit('Golden spiral cannot be generated', 2)
            return False

        self.processPoints()

        return True

    def genBuildSpiralMed(self):
        """

        :return: success
        """

        self.lastGenerator = 'spiralMed'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.generateGoldenSpiral(numberPoints=250, keep=keep)

        if not suc:
            self.app.message.emit('Golden spiral cannot be generated', 2)
            return False

        self.processPoints()

        return True

    def genBuildSpiralNorm(self):
        """

        :return: success
        """

        self.lastGenerator = 'spiralNorm'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.generateGoldenSpiral(numberPoints=150, keep=keep)

        if not suc:
            self.app.message.emit('Golden spiral cannot be generated', 2)
            return False

        self.processPoints()

        return True

    def genBuildSpiralMin(self):
        """

        :return: success
        """

        self.lastGenerator = 'spiralMin'
        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.generateGoldenSpiral(numberPoints=75, keep=keep)

        if not suc:
            self.app.message.emit('Golden spiral cannot be generated', 2)
            return False

        self.processPoints()

        return True

    def genBuildFile(self):
        """
        genBuildFile tries to load a give build point file and displays it for usage.

        :return: success
        """

        self.lastGenerator = 'file'
        fileName = self.ui.buildPFileName.text()
        if not fileName:
            self.app.message.emit('Build points file name not given', 2)
            return False

        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.loadBuildP(fileName=fileName, keep=keep)

        if not suc:
            text = 'Build points file [{0}] could not be loaded'.format(fileName)
            self.app.message.emit(text, 2)
            return False

        self.processPoints()

        return True

    def loadBuildFile(self):
        """
        loadBuildFile calls a file selector box and selects the filename to be loaded

        :return: success
        """

        folder = self.app.mwGlob['configDir']
        fileTypes = 'Build Point Files (*.bpts);; CSV Files (*.csv);; MW3 Files (*.txt)'
        loadFilePath, fileName, ext = self.openFile(self,
                                                    'Open build point file',
                                                    folder,
                                                    fileTypes)
        if not loadFilePath:
            return False

        keep = self.ui.keepGeneratedPoints.isChecked()
        suc = self.app.data.loadBuildP(fileName=fileName, ext=ext, keep=keep)

        if suc:
            self.ui.buildPFileName.setText(fileName)
            self.app.message.emit(f'Build file [{fileName}] loaded', 0)
        else:
            self.app.message.emit(f'Build file [{fileName}] cannot no be loaded', 2)

        self.genBuildFile()

        return True

    def saveBuildFile(self):
        """
        saveBuildFile calls saving the build file

        :return: success
        """

        fileName = self.ui.buildPFileName.text()
        if not fileName:
            self.app.message.emit('Build points file name not given', 2)
            return False

        suc = self.app.data.saveBuildP(fileName=fileName)

        if suc:
            self.app.message.emit(f'Build file [{fileName}] saved', 0)
        else:
            self.app.message.emit(f'Build file [{fileName}] cannot no be saved', 2)

        return True

    def saveBuildFileAs(self):
        """
        saveBuildFileAs calls a file selector box and selects the filename to be save

        :return: success
        """

        folder = self.app.mwGlob['configDir']
        saveFilePath, fileName, ext = self.saveFile(self,
                                                    'Save build point file',
                                                    folder,
                                                    'Build point files (*.bpts)',
                                                    )
        if not saveFilePath:
            return False

        suc = self.app.data.saveBuildP(fileName=fileName)

        if suc:
            self.ui.buildPFileName.setText(fileName)
            self.app.message.emit(f'Build file [{fileName}] saved', 0)
        else:
            self.app.message.emit(f'Build file [{fileName}] cannot no be saved', 2)

        return True

    def clearBuildP(self):
        """

        :return: success
        """

        self.app.data.clearBuildP()
        self.app.drawBuildPoints.emit()

        if not self.app.uiWindows['showHemisphereW']['classObj']:
            return False

        self.app.uiWindows['showHemisphereW']['classObj'].clearHemisphere()

        return True

    def autoDeletePoints(self):
        """
        autoDeletePoints removes all generated or visible build points below the horizon line
        or within the limits of the meridian flip and redraws the hemisphere window.

        :return: True for test purpose
        """

        if self.ui.checkAutoDeleteHorizon.isChecked():
            self.app.data.deleteBelowHorizon()

        if self.ui.checkAutoDeleteMeridian.isChecked():
            self.app.data.deleteCloseMeridian()

        if self.ui.checkSafetyMarginHorizon.isChecked():
            value = self.ui.safetyMarginHorizon.value()
            self.app.data.deleteCloseHorizonLine(value)

        return True

    def autoSortPoints(self):
        """
        autoSortPoints sort the given build point first to east and west and than based
        on the decision high altitude to low altitude or east to west in each hemisphere

        :return: success if sorted
        """

        eastwest = self.ui.checkSortEW.isChecked()
        highlow = self.ui.checkSortHL.isChecked()
        avoidFlip = self.ui.checkAvoidFlip.isChecked()

        pierside = self.app.mount.obsSite.pierside

        if pierside is None:
            avoidFlip = False

        if not eastwest and not highlow and not avoidFlip:
            return False

        if avoidFlip:
            self.app.data.sort(eastwest=eastwest,
                               highlow=highlow,
                               pierside=pierside)
        else:
            self.app.data.sort(eastwest=eastwest, highlow=highlow)

        return True

    def processPoints(self):
        """

        :return: True for test purpose
        """

        self.autoDeletePoints()
        self.autoSortPoints()
        self.app.redrawHemisphere.emit()
        self.app.drawBuildPoints.emit()

        return True



############################################################
# -*- coding: utf-8 -*-
#
#       #   #  #   #   #    #
#      ##  ##  #  ##  #    #
#     # # # #  # # # #    #  #
#    #  ##  #  ##  ##    ######
#   #   #   #  #   #       #
#
# Python-based Tool for interaction with the 10micron mounts
# GUI with PyQT5 for python
#
# written in python3, (c) 2019-2021 by mworion
# Licence APL2.0
#
###########################################################
# standard libraries

# external packages

# local import


class Relay(object):
    """
    """

    def __init__(self):
        self.app.relay.signals.statusReady.connect(self.updateRelayGui)

    def initConfig(self):
        # config = self.app.config['mainW']
        return True

    def storeConfig(self):
        # config = self.app.config['mainW']
        return True

    def updateRelayGui(self):
        """
        updateRelayGui changes the style of the button related to the state of the relay

        :return: success for test
        """

        for status, button in zip(self.app.relay.status, self.relayButtons):
            if status:
                self.changeStyleDynamic(button, 'running', 'true')
            else:
                self.changeStyleDynamic(button, 'running', 'false')
        return True
